import sys
from pathlib import Path

from hatchling.metadata.plugin.interface import MetadataHookInterface

# Building version-pioneer using version-pioneer functions..
# without installing the package, we import the module.
# Make sure build-time dependencies are installed.
sys.path.append(str(Path(__file__).parent / "src"))

from version_pioneer.build.hatchling.build_hook import (
    VersionPioneerBuildHook,  # This hook will be used just by importing it.  # noqa: F401
)
from version_pioneer.utils.version_script import (
    exec_version_script,
    find_version_script_from_project_dir,
)


# We can't use VersionSource plugin as custom, so we use MetadataHook.
class CustomMetadataHook(MetadataHookInterface):
    def update(self, metadata: dict) -> None:
        """
        This updates the metadata mapping of the `project` table in-place.
        """
        # This also checks the valid config, so run it first.
        versionscript = find_version_script_from_project_dir(
            project_dir=self.root,
            either_versionfile_or_versionscript=True,
        )
        version_dict = exec_version_script(versionscript)
        metadata["version"] = version_dict["version"]



from __future__ import annotations

import logging
import re
from pathlib import Path

from version_pioneer.utils.version_script import (
    exec_version_script,
    exec_version_script_code,
)

logger = logging.getLogger(__name__)


class VersionScriptResolutionError(Exception):
    pass


def verify_resolved_versionfile(resolved_versionscript_code: str):
    # does it have `def get_version_dict():`?
    if (
        re.search(
            r"^def get_version_dict\(.*\):$", resolved_versionscript_code, re.MULTILINE
        )
        is None
    ):
        raise VersionScriptResolutionError(
            f"Resolved _version.py code does not contain `def get_version_dict(): ...`: {resolved_versionscript_code}"
        )

    # Can you execute it without dependencies?
    version_dict = exec_version_script_code(resolved_versionscript_code)

    # Can you get the version?
    version = version_dict["version"]
    assert isinstance(version, str)

    # Does it have all keys in VersionDict TypedDict definition?
    # In TypedDict, __required_keys__, __optional_keys__ added in python 3.9
    # get_annotations() added in python 3.10
    # With lazy evaluation of ForwardRef, this is tricky.

    # for key, type_ in VersionDict.__annotations__.items():
    #     assert key in version_dict
    #     if isinstance(type_, ForwardRef):
    #         # https://stackoverflow.com/questions/76106117/python-resolve-forwardref
    #         # type_ = type_._evaluate(globals(), locals(), frozenset())  # Python 3.9+
    #         type_ = type_._evaluate(globals(), locals())
    #     assert isinstance(version_dict[key], type_)


def get_dynamic_version(project_dir: Path) -> str:
    version_module_code = project_dir / "src" / "my_app" / "_version.py"
    return exec_version_script(version_module_code)["version"]



from __future__ import annotations

import filecmp
import fnmatch
import os
from collections.abc import Iterable, Sequence
from os import PathLike
from pathlib import Path


def remove_files_recusively(directory: str | PathLike, patterns: str | Sequence[str]):
    """Remove files recursively from a directory that match any of the patterns."""
    patterns = [patterns] if isinstance(patterns, str) else patterns

    for root, _, files in os.walk(directory):
        root = Path(root)
        for file in files:
            if any(fnmatch.fnmatch(str(root / file), pattern) for pattern in patterns):
                (root / file).unlink()


def are_dir_trees_equal(dir1, dir2):
    """
    Compare two directories recursively. Files in each directory are
    assumed to be equal if their names and contents are equal.
    """
    dirs_cmp = filecmp.dircmp(dir1, dir2)
    if (
        len(dirs_cmp.left_only) > 0
        or len(dirs_cmp.right_only) > 0
        or len(dirs_cmp.funny_files) > 0
    ):
        raise FileNotFoundError(
            f"Directory trees are not equal: {dirs_cmp.left_only}, {dirs_cmp.right_only}, {dirs_cmp.funny_files}"
        )
    (_, mismatch, errors) = filecmp.cmpfiles(
        dir1, dir2, dirs_cmp.common_files, shallow=False
    )
    if len(mismatch) > 0 or len(errors) > 0:
        raise FileNotFoundError(f"Directory trees are not equal: {mismatch}, {errors}")
    for common_dir in dirs_cmp.common_dirs:
        new_dir1 = os.path.join(dir1, common_dir)  # noqa: PTH118
        new_dir2 = os.path.join(dir2, common_dir)  # noqa: PTH118
        are_dir_trees_equal(new_dir1, new_dir2)


def find_root_dir_with_file(
    source: str | PathLike, marker: str | Iterable[str]
) -> Path:
    """
    Find the first parent directory containing a specific "marker", relative to a file path.
    """
    source = Path(source).resolve()
    if isinstance(marker, str):
        marker = {marker}

    while source != source.parent:
        if any((source / m).exists() for m in marker):
            return source

        source = source.parent

    raise FileNotFoundError(f"File {marker} not found in any parent directory")



from __future__ import annotations

import json
import logging
from enum import Enum
from os import PathLike
from pathlib import Path
from typing import Any, Literal, TypeVar

from version_pioneer import template
from version_pioneer.utils.toml import (
    find_pyproject_toml,
    get_toml_value,
    load_toml,
)
from version_pioneer.versionscript import VersionDict

logger = logging.getLogger(__name__)


class ResolutionFormat(str, Enum):
    python = "python"
    json = "json"
    version_string = "version-string"


RESOLUTION_FORMAT_TYPE = TypeVar(
    "RESOLUTION_FORMAT_TYPE",
    Literal["python", "json", "version-string"],
    ResolutionFormat,
)


def find_version_script_from_pyproject_toml_dict(
    pyproject_toml_dict: dict[str, Any],
    *,
    either_versionfile_or_versionscript: bool = True,
):
    versionscript: Path | None = get_toml_value(
        pyproject_toml_dict,
        ["tool", "version-pioneer", "versionscript"],
        return_path_object=True,
    )

    if versionscript is None:
        # NOTE: even if we end up loading versionfile-sdist, we still need to check the valid config.
        raise KeyError(
            "Missing key tool.version-pioneer.versionscript in pyproject.toml"
        )

    if either_versionfile_or_versionscript:
        versionfile: Path | None = get_toml_value(
            pyproject_toml_dict,
            ["tool", "version-pioneer", "versionfile-sdist"],
            return_path_object=True,
        )
        if versionfile is not None and versionfile.exists():
            return versionfile

    if not versionscript.exists():
        raise FileNotFoundError(f"Version script not found: {versionscript}")

    return versionscript


def find_version_script_from_project_dir(
    project_dir: str | PathLike | None = None,
    *,
    either_versionfile_or_versionscript: bool = True,
):
    """
    Args:
        either_versionfile_or_versionscript: If True, return either versionfile-sdist if it exists,
            else versionscript.
            This is important because in sdist build, the versionfile is already evaluated
            and git tags are not available.
    """
    if project_dir is None:
        project_dir = Path.cwd()
    else:
        project_dir = Path(project_dir)

    if project_dir.is_file():
        raise NotADirectoryError(f"{project_dir} is not a directory.")

    pyproject_toml_file = find_pyproject_toml(project_dir)
    pyproject_toml = load_toml(pyproject_toml_file)

    return pyproject_toml_file.parent / find_version_script_from_pyproject_toml_dict(
        pyproject_toml,
        either_versionfile_or_versionscript=either_versionfile_or_versionscript,
    )


def exec_version_script_code(version_script_code: str) -> VersionDict:
    """
    Execute `get_version_dict()` in _version.py.
    """
    module_globals = {}
    exec(version_script_code, module_globals)
    return module_globals["get_version_dict"]()


def exec_version_script(
    version_script_path: str | PathLike,
) -> VersionDict:
    """Execute _version.py to get __version_dict__."""
    version_script_path = Path(version_script_path)
    code = version_script_path.read_text()
    return exec_version_script_code(code)


def convert_version_dict(
    version_dict: VersionDict,
    output_format: RESOLUTION_FORMAT_TYPE,
) -> str:
    from version_pioneer import __version__

    if output_format == ResolutionFormat.python:
        return template.EXEC_OUTPUT_PYTHON.format(
            version_pioneer_version=__version__,
            version_dict=version_dict,
        )
    elif output_format == ResolutionFormat.json:
        return json.dumps(version_dict)
    elif output_format == ResolutionFormat.version_string:
        return version_dict["version"]
    else:
        raise ValueError(f"Invalid output format: {output_format}")



from __future__ import annotations

import sys
from os import PathLike
from pathlib import Path
from typing import Any

if sys.version_info >= (3, 11):
    import tomllib
else:
    import tomli as tomllib


def load_toml(file: str | PathLike) -> dict[str, Any]:
    with open(file, "rb") as f:
        return tomllib.load(f)


def get_toml_value(
    toml_dict: dict[str, Any],
    keys: list[str],
    *,
    default: Any = None,
    raise_error: bool = False,
    toml_path_for_error: str = "pyproject.toml",
    return_path_object: bool = False,
) -> Any:
    """
    Args:
        return_path_object: Return a path object safely (when returning a default, don't convert to Path).
    """
    if len(keys) == 0:
        raise ValueError("keys must not be empty")

    if default is not None and raise_error:
        raise ValueError("default and raise_error cannot both be set.")

    for k in keys:
        if k not in toml_dict:
            key = ".".join(keys)
            if raise_error:
                raise KeyError(f"Missing key {key} in {toml_path_for_error}")
            return default
        toml_dict = toml_dict[k]

    value: Any = toml_dict

    if return_path_object:
        return Path(value)
    return value


def find_pyproject_toml(project_dir: str | PathLike | None = None) -> Path:
    """
    Find the pyproject.toml file in the current directory or any parent directory.
    """
    from .files import find_root_dir_with_file

    if project_dir is None:
        project_dir = Path.cwd()
    return find_root_dir_with_file(project_dir, "pyproject.toml") / "pyproject.toml"



import os
import subprocess
import textwrap
from pathlib import Path
from tempfile import TemporaryDirectory

import pytest

from version_pioneer.api import get_version_script_core_code
from version_pioneer.template import SETUP_PY
from version_pioneer.utils.build import build_project

SCRIPT_DIR = Path(__file__).resolve().parent
os.environ["GIT_CONFIG_GLOBAL"] = str(SCRIPT_DIR / "gitconfig")


# @pytest.fixture(name="plugin_dir", scope="session")
# def _plugin_dir():
#     """
#     Install the plugin into a temporary directory with a random path to
#     prevent pip from caching it.
#
#     Copy only the src directory, pyproject.toml, and whatever is needed
#     to build ourselves.
#     """
#     with TemporaryDirectory() as d:
#         directory = Path(d, "plugin")
#         shutil.copytree(Path.cwd() / "src", directory / "src")
#         shutil.copytree(Path.cwd() / "deps", directory / "deps")
#         # required because this plugin uses git to get version
#         shutil.copytree(Path.cwd() / ".git", directory / ".git")
#         for fn in [
#             "pyproject.toml",
#             "LICENSE",
#             "README.md",
#             "hatch_build.py",
#         ]:
#             shutil.copy(Path.cwd() / fn, directory / fn)
#
#         yield directory.resolve()


@pytest.fixture(name="plugin_wheel", scope="session")
def _plugin_wheel():
    """
    Build the plugin into a temporary directory with a random path to
    prevent pip from caching it.

    Instead of above approach that copies the source code, building the plugin
    ensures that the plugin is built correctly for deployment.

    For example, it ensures _version.py and _version_orig.py are included in the wheel with different content.
    """
    with TemporaryDirectory() as d:
        out, builds = build_project("--out-dir", d)
        assert len(builds) == 2
        wheel_path = builds[1]
        if not wheel_path.is_absolute():
            # pyproject-build does not print absolute path
            wheel_path = Path(d) / wheel_path

        yield wheel_path


@pytest.fixture(name="new_hatchling_project")
def _new_hatchling_project(plugin_wheel: Path, tmp_path: Path, monkeypatch):
    """
    Create, and cd into, a blank new project that is configured to use our temporary plugin installation.
    """
    project_dir = tmp_path / "my-app"
    project_dir.mkdir()

    pyproject_file = project_dir / "pyproject.toml"
    pyproject_file.write_text(
        textwrap.dedent(f"""
            [build-system]
            requires = ["hatchling", "version-pioneer @ {plugin_wheel.as_uri()}"]
            build-backend = "hatchling.build"

            [tool.hatch.version]
            source = "version-pioneer"

            [tool.hatch.build.hooks.version-pioneer]

            [tool.version-pioneer]
            versionscript = "src/my_app/_version.py"
            versionfile-sdist = "src/my_app/_version.py"
            versionfile-wheel = "my_app/_version.py"

            [project]
            name = "my-app"
            dynamic = ["version"]
            requires-python = ">=3.8"
        """),
        encoding="utf-8",
    )

    package_dir = project_dir / "src" / "my_app"
    package_dir.mkdir(parents=True)

    # NOTE: without gitignore, build will create artifacts which makes the version always dirty,
    # and include wrong files in the package
    gitignore_file = project_dir / ".gitignore"
    gitignore_file.write_text(
        textwrap.dedent("""
            /dist/
            /dist-*/
        """)
    )

    package_root = package_dir / "__init__.py"
    package_root.write_text("")

    version_file = package_dir / "_version.py"
    version_file.write_text(get_version_script_core_code())

    monkeypatch.chdir(project_dir)

    assert Path.cwd() == project_dir

    subprocess.run(["git", "init"], check=True)
    subprocess.run(["git", "add", "."], check=True)
    subprocess.run(["git", "commit", "-m", "Initial commit"], check=True)
    subprocess.run(["git", "tag", "v0.1.0"], check=True)

    return project_dir


@pytest.fixture(name="new_setuptools_project")
def _new_setuptools_project(plugin_wheel: Path, tmp_path: Path, monkeypatch):
    """
    Create, and cd into, a blank new project that is configured to use our temporary plugin installation.
    """
    project_dir = tmp_path / "my-app"
    project_dir.mkdir()

    pyproject_file = project_dir / "pyproject.toml"
    pyproject_file.write_text(
        textwrap.dedent(f"""
            [build-system]
            requires = ["setuptools", "version-pioneer @ {plugin_wheel.as_uri()}"]
            build-backend = "setuptools.build_meta"

            [tool.version-pioneer]
            versionscript = "src/my_app/_version.py"
            versionfile-sdist = "src/my_app/_version.py"
            versionfile-wheel = "my_app/_version.py"

            [project]
            name = "my-app"
            dynamic = ["version"]
            requires-python = ">=3.8"
        """),
        encoding="utf-8",
    )

    setup_file = project_dir / "setup.py"
    setup_file.write_text(SETUP_PY)

    # NOTE: without gitignore, build will create artifacts which makes the version always dirty
    gitignore_file = project_dir / ".gitignore"
    gitignore_file.write_text(
        textwrap.dedent("""
            *.egg-info/
            /dist/
            /dist-*/
        """)
    )

    package_dir = project_dir / "src" / "my_app"
    package_dir.mkdir(parents=True)

    package_root = package_dir / "__init__.py"
    package_root.write_text("")

    version_file = package_dir / "_version.py"
    version_file.write_text(get_version_script_core_code())

    monkeypatch.chdir(project_dir)

    assert Path.cwd() == project_dir

    subprocess.run(["git", "init"], check=True)
    subprocess.run(["git", "add", "."], check=True)
    subprocess.run(["git", "commit", "-m", "Initial commit"], check=True)
    subprocess.run(["git", "tag", "v0.1.0"], check=True)

    return project_dir


@pytest.fixture(name="new_pdm_project")
def _new_pdm_project(plugin_wheel: Path, tmp_path: Path, monkeypatch):
    """
    Create, and cd into, a blank new project that is configured to use our temporary plugin installation.
    """
    project_dir = tmp_path / "my-app"
    project_dir.mkdir()

    pyproject_file = project_dir / "pyproject.toml"
    pyproject_file.write_text(
        textwrap.dedent(f"""
            [build-system]
            requires = ["pdm-backend", "version-pioneer @ {plugin_wheel.as_uri()}"]
            build-backend = "pdm.backend"

            [tool.version-pioneer]
            versionscript = "src/my_app/_version.py"
            versionfile-sdist = "src/my_app/_version.py"
            versionfile-wheel = "my_app/_version.py"

            [project]
            name = "my-app"
            dynamic = ["version"]
            requires-python = ">=3.8"
        """),
        encoding="utf-8",
    )

    # NOTE: without gitignore, build will create artifacts which makes the version always dirty
    gitignore_file = project_dir / ".gitignore"
    gitignore_file.write_text(
        textwrap.dedent("""
            /dist/
            /dist-*/
        """)
    )

    package_dir = project_dir / "src" / "my_app"
    package_dir.mkdir(parents=True)

    package_root = package_dir / "__init__.py"
    package_root.write_text("")

    version_file = package_dir / "_version.py"
    version_file.write_text(get_version_script_core_code())

    monkeypatch.chdir(project_dir)

    assert Path.cwd() == project_dir

    subprocess.run(["git", "init"], check=True)
    subprocess.run(["git", "add", "."], check=True)
    subprocess.run(["git", "commit", "-m", "Initial commit"], check=True)
    subprocess.run(["git", "tag", "v0.1.0"], check=True)

    return project_dir



#!/usr/bin/env python3
"""
Generate a version number from Git tags (e.g. tag "v1.2.3" and 4 commits -> "1.2.3+4.g123abcdef" in "pep440" style).

This "versionscript" may be replaced with a much shorter "versionfile" in distribution tarballs
(built by `uv build`, `pyproject-build`) that just contains one method:
`def get_version_dict() -> VersionDict: return {"version": "0.1.0", ...}`.

Refactored from Versioneer's _version.py (for git).

Note:
    - Should be compatible with python 3.8+ without any dependencies.
        - (For dev) Avoid importing third-party libraries, including version-pioneer itself
          because this file can get vendored into other projects.
        - (For user) Once the script is vendored, and you want to customise it, of course you can
          import other libraries and add those in build-time dependencies.
    - This file is usually located at `src/my_package/_version.py`.
    - `src/my_package/__init__.py` should define `__version__ = get_version_dict()["version"]` by importing this module.
    - It should also be able to be run as a script to print the version info in json format.
    - It is often `exec`-uted and `get_version_dict()` is evaluated from this file.
        - Using `from __future__ import ...` with dataclasses makes it hard to "exec" this file,
          so you MUST NOT use both here.
            - See https://github.com/mkdocs/mkdocs/issues/3141
              https://github.com/pypa/hatch/issues/1863
              https://github.com/sqlalchemy/alembic/issues/1419
            - You need to put the module in `sys.modules` before executing
              because dataclass will look for the type there.
            - While this can be fixed, it's a common gotcha and I expect that
              some build backends or tools will be buggy.
            - For now, we don't future import. In dataclass definition,
              we use `typing.Optional` instead of `| None`
              until we drop support for Python 3.9.
"""

# ruff: noqa: T201 FA100

import contextlib
import errno
import functools
import os
import re
import subprocess
import sys
from collections.abc import Iterable
from dataclasses import dataclass
from email.parser import Parser
from enum import Enum
from os import PathLike
from pathlib import Path
from typing import Any, Literal, Optional, TypedDict, TypeVar, Union


class VersionStyle(str, Enum):
    pep440 = "pep440"
    pep440_master = "pep440-master"
    pep440_branch = "pep440-branch"
    pep440_pre = "pep440-pre"
    pep440_post = "pep440-post"
    pep440_post_branch = "pep440-post-branch"
    git_describe = "git-describe"
    git_describe_long = "git-describe-long"
    digits = "digits"


VERSION_STYLE_TYPE = TypeVar(
    "VERSION_STYLE_TYPE",
    Literal[
        "pep440",
        "pep440-master",
        "pep440-branch",
        "pep440-pre",
        "pep440-post",
        "pep440-post-branch",
        "git-describe",
        "git-describe-long",
        "digits",
    ],
    VersionStyle,
)


# ┌──────────────────────────────────────────┐
# │     Modify the configuration below.      │
# └──────────────────────────────────────────┘
@dataclass(frozen=True)
class VersionPioneerConfig:
    style: VersionStyle = VersionStyle.pep440
    tag_prefix: str = "v"
    # if there is no .git, like it's a source tarball downloaded from GitHub Releases,
    # find version from the name of the parent directory.
    # e.g. setting it to "github-repo-name-" will find the version from "github-repo-name-1.2.3"
    # Set it to None to try to determine the prefix from pyproject.toml.
    parentdir_prefix: Optional[str] = None
    verbose: bool = False


# ┌──────────────────────────────────────────┐
# │     Modify the configuration above.      │
# └──────────────────────────────────────────┘


class VersionDict(TypedDict):
    """Return type of get_version_dict()."""

    version: str
    full_revisionid: "str | None"
    dirty: "bool | None"
    error: "str | None"
    date: "str | None"


try:
    _SCRIPT_DIR_OR_CURRENT_DIR = Path(__file__).resolve().parent
except NameError:
    # NOTE: py2exe/bbfreeze/non-cpython implementations may not have __file__.
    # and usually during installation when this file is evaluated, __file__ doesn't exist.
    # However, once you installed in editable mode (e.g. `pip install -e .`),
    # __file__ will be available and more reliable.
    _SCRIPT_DIR_OR_CURRENT_DIR = Path.cwd()


MASTER_BRANCHES = ("master", "main")

if sys.platform == "win32":
    GIT_COMMANDS = ["git.cmd", "git.exe"]
else:
    GIT_COMMANDS = ["git"]

# GIT_DIR can interfere with correct operation of Versioneer.
# It may be intended to be passed to the Versioneer-versioned project,
# but that should not change where we get our version from.
env = os.environ.copy()
env.pop("GIT_DIR", None)


class NotThisMethodError(Exception):
    """Exception raised if a method is not valid for the current scenario."""


class CurrentBranchIsMasterError(Exception):
    """GitMasterDistance can't be initialised if the current branch is master."""


@dataclass(frozen=True)
class GitMasterDistance:
    """
    Compute the distance from the tag until master, and from master to current branch.

    Useful if you often create a develop branch from master.

    Relevant commands:
        ```console
        $ git branch --show-current
        feat/branch-name

        $ # use with -a to include remote branches
        $ git branch --contains <commit>
        * master
          feat/branch-name

        $ # Notice that there are no master (only the origin/master) in the list
        $ # because the master has more commits than the current branch.
        $ # Thus we can't use the '%D' ref names.
        $ # We instead use `git branch --contains` on each commit to find the branch.
        $ git log v0.3.2..[BRANCH] --pretty=format:"%H,%h,%D"  # BRANCH is optional (default: current branch)
        87e38450d1fce0398fbc9de08f2abe3e5da0431e,87e3845,trash/version
        8a76eac0f107dd3810c3cfd5c89e92c7f5d31e50,8a76eac,
        a855d640912f728b8946eddba41ed5b2a992f394,a855d64,origin/master, origin/HEAD
        5cb3c6663f494b3c99dfd23b5394fa2da7f49cef,5cb3c66,
        2c00c0ed4e46a5459bc70f47739a0c50a789d3c3,2c00c0e,
        812d3e29666f2d75b80b4160532fa25afaab2ffd,812d3e2,tag: v0.3.3
        2127fd373d14ed5ded497fc18ac1c1b667f93a7d,2127fd3,
        ae7cb503342d551e2503dc0a90be656946342743,ae7cb50,
        ```

    References:
        - https://stackoverflow.com/questions/4649356/how-do-i-run-git-log-to-see-changes-only-for-a-specific-branch
        - https://stackoverflow.com/questions/2706797/finding-what-branch-a-git-commit-came-from
        - https://stackoverflow.com/questions/3998883/git-how-to-find-commit-hash-where-branch-originated-from
            - The reason we can't use reflog..
    """

    current_branch: str
    distance_from_tag_to_master: int
    distance_from_master: int
    master_commit: str

    @property
    def master_commit_short(self) -> "str | None":
        """Return the short commit hash of the master commit."""
        if self.master_commit is not None:
            return self.master_commit[:7]
        return None

    @classmethod
    def from_git(
        cls: "type[GitMasterDistance]",
        tag_of_interest: "str | None",
        *,
        cwd: "str | PathLike",
        verbose: bool = False,
    ) -> "GitMasterDistance":
        git_runner = functools.partial(
            _run_git_command_or_error, env=env, verbose=verbose
        )

        # Get the current branch name
        current_branch, rc = git_runner(
            ["branch", "--show-current"],
            cwd=cwd,
        )
        current_branch = current_branch.strip()

        if current_branch in MASTER_BRANCHES:
            raise CurrentBranchIsMasterError(
                "Current branch is master. Can't compute distance to/from master."
            )

        if tag_of_interest is None:
            # Get entire history (commits) from the current branch
            out, rc = git_runner(
                ["log", "--pretty=format:%H"],
                cwd=cwd,
            )
        else:
            # Get history from the tag to the current branch
            out, rc = git_runner(
                ["log", f"{tag_of_interest}..", "--pretty=format:%H"],
                cwd=cwd,
            )

        all_commits = out.splitlines()

        # NOTE: all commits could be no output if the tag is the same as the current branch.

        # Search from the top, and find the first commit that shares the master branch
        distance_from_master = 0
        master_commit = None
        for commit in all_commits:
            out, rc = git_runner(
                ["branch", "--contains", commit],
                cwd=cwd,
            )
            branches = out.splitlines()
            # Strip off the leading "* " from the list of branches.
            branches = [branch.lstrip("* ") for branch in branches]
            if any(
                master_branch_name in branches for master_branch_name in MASTER_BRANCHES
            ):
                master_commit = commit
                break

            distance_from_master += 1

        if master_commit is None:
            # Can't find master? We assume that it's the tag. It may not always be. (like release branch)
            if tag_of_interest is None:
                if verbose:
                    print(
                        "No tag found and none of the commit history points to master/main."
                    )
                    print("Maybe detached head or you don't use master?")
                raise NotThisMethodError(
                    "No tag found and none of the commit history points to master/main. "
                    "Maybe detached head or you don't use master?"
                )

            out, re = git_runner(["rev-list", "-1", tag_of_interest], cwd=cwd)
            master_commit = out.strip()
            if len(master_commit) != 40:
                raise NotThisMethodError("Something is strange in you git commit hash")

        return cls(
            current_branch=current_branch,
            distance_from_tag_to_master=len(all_commits) - distance_from_master,
            distance_from_master=distance_from_master,
            master_commit=master_commit,
        )


@dataclass(frozen=True)
class GitPieces:
    """
    Get version from 'git describe' in the root of the source tree.

    This only gets called if _version.py hasn't already been rewritten with a short
    version string, meaning we're inside a checked out source tree.
    """

    long: str
    short: str
    branch: str
    dirty: bool

    # options to pass to GitMasterDistance
    cwd: Union[str, PathLike]
    verbose: bool

    error: Optional[str] = None
    distance: Optional[int] = None
    closest_fulltag: Optional[str] = None  # include tag_prefix (v1.0.0)
    closest_tag: Optional[str] = None  # strip tag_prefix (1.0.0)
    date: Optional[str] = None

    @classmethod
    def from_git(
        cls: "type[GitPieces]",
        tag_prefix: str,
        *,
        cwd: "str | PathLike",
        verbose: bool = False,
    ) -> "GitPieces":
        runner = functools.partial(_run_command, env=env, verbose=verbose)

        _, rc = runner(
            GIT_COMMANDS, ["rev-parse", "--git-dir"], cwd=cwd, hide_stderr=not verbose
        )
        if rc != 0:
            if verbose:
                print(f"Directory {cwd} not under git control")
            raise NotThisMethodError("'git rev-parse --git-dir' returned error")

        # if there is a tag matching tag_prefix, this yields TAG-NUM-gHEX[-dirty]
        # if there isn't one, this yields HEX[-dirty] (no NUM)
        describe_out, rc = runner(
            GIT_COMMANDS,
            [
                "describe",
                "--tags",
                "--dirty",
                "--always",
                "--long",
                "--match",
                f"{tag_prefix}[[:digit:]]*",
            ],
            cwd=cwd,
        )
        # --long was added in git-1.5.5
        if describe_out is None:
            raise NotThisMethodError("'git describe' failed")
        describe_out = describe_out.strip()
        full_out, rc = runner(GIT_COMMANDS, ["rev-parse", "HEAD"], cwd=cwd)
        if full_out is None:
            raise NotThisMethodError("'git rev-parse' failed")
        full_out = full_out.strip()

        pieces: dict[str, Any] = {}
        pieces["long"] = full_out
        pieces["short"] = full_out[:7]  # maybe improved later
        pieces["error"] = None

        branch_name, rc = runner(
            GIT_COMMANDS, ["rev-parse", "--abbrev-ref", "HEAD"], cwd=cwd
        )
        # --abbrev-ref was added in git-1.6.3
        if rc != 0 or branch_name is None:
            raise NotThisMethodError("'git rev-parse --abbrev-ref' returned error")
        branch_name = branch_name.strip()

        if branch_name == "HEAD":
            # If we aren't exactly on a branch, pick a branch which represents
            # the current commit. If all else fails, we are on a branchless
            # commit.
            branches, rc = runner(GIT_COMMANDS, ["branch", "--contains"], cwd=cwd)
            # --contains was added in git-1.5.4
            if rc != 0 or branches is None:
                raise NotThisMethodError("'git branch --contains' returned error")
            branches = branches.split("\n")

            # Remove the first line if we're running detached
            if "(" in branches[0]:
                branches.pop(0)

            # Strip off the leading "* " from the list of branches.
            branches = [branch.lstrip("* ") for branch in branches]
            if not branches:
                branch_name = None
            else:
                for master_branch_name in MASTER_BRANCHES:
                    if master_branch_name in branches:
                        branch_name = master_branch_name
                        break
                else:
                    # Pick the first branch that is returned. Good or bad.
                    branch_name = branches[0]

        pieces["branch"] = branch_name

        # parse describe_out. It will be like TAG-NUM-gHEX[-dirty] or HEX[-dirty]
        # TAG might have hyphens.
        git_describe = describe_out

        # look for -dirty suffix
        dirty = git_describe.endswith("-dirty")
        pieces["dirty"] = dirty
        if dirty:
            git_describe = git_describe[: git_describe.rindex("-dirty")]

        # now we have TAG-NUM-gHEX or HEX

        if "-" in git_describe:
            # TAG-NUM-gHEX
            mo = re.search(r"^(.+)-(\d+)-g([0-9a-f]+)$", git_describe)
            if not mo:
                # unparsable. Maybe git-describe is misbehaving?
                pieces["error"] = (
                    f"unable to parse git-describe output: '{describe_out}'"
                )
                return cls(**pieces, cwd=cwd, verbose=verbose)

            # tag
            full_tag = mo.group(1)
            if not full_tag.startswith(tag_prefix):
                if verbose:
                    print(f"tag '{full_tag}' doesn't start with prefix '{tag_prefix}'")
                pieces["error"] = (
                    f"tag '{full_tag}' doesn't start with prefix '{tag_prefix}'"
                )
                return cls(**pieces, cwd=cwd, verbose=verbose)
            pieces["closest_fulltag"] = full_tag
            pieces["closest_tag"] = full_tag[len(tag_prefix) :]

            # distance: number of commits since tag
            pieces["distance"] = int(mo.group(2))

            # commit: short hex revision ID
            pieces["short"] = mo.group(3)

        else:
            # HEX: no tags
            pieces["closest_fulltag"] = None
            pieces["closest_tag"] = None
            out, rc = runner(
                GIT_COMMANDS, ["rev-list", "HEAD", "--left-right"], cwd=cwd
            )
            assert out is not None
            pieces["distance"] = len(out.split())  # total number of commits

        # commit date: see ISO-8601 comment in git_versions_from_keywords()
        out, rc = runner(GIT_COMMANDS, ["show", "-s", "--format=%ci", "HEAD"], cwd=cwd)
        assert out is not None
        date = out.strip()
        # Use only the last line.  Previous lines may contain GPG signature
        # information.
        date = date.splitlines()[-1]
        pieces["date"] = date.strip().replace(" ", "T", 1).replace(" ", "", 1)

        return cls(**pieces, cwd=cwd, verbose=verbose)

    @property
    def _plus_or_dot(self) -> str:
        """Return a + if we don't already have one, else return a ."""
        if self.closest_tag is None:
            return "+"
        elif "+" in self.closest_tag:
            return "."
        return "+"

    def _render_pep440(self) -> str:
        """
        Build up version string, with post-release "local version identifier".

        Our goal: TAG[+DISTANCE.gHEX[.dirty]] . Note that if you
        get a tagged build and then dirty it, you'll get TAG+0.gHEX.dirty

        Exceptions:
        1: no tags. git_describe was just HEX. 0+untagged.DISTANCE.gHEX[.dirty]
        """
        if self.closest_tag:
            rendered = self.closest_tag
            if self.distance or self.dirty:
                rendered += self._plus_or_dot
                rendered += f"{self.distance}.g{self.short}"
                if self.dirty:
                    rendered += ".dirty"
        else:
            # exception #1
            rendered = f"0+untagged.{self.distance}.g{self.short}"
            if self.dirty:
                rendered += ".dirty"
        return rendered

    def _render_pep440_master(self) -> str:
        """
        Render including master branch distance and commit like TAG[+MASTERDIST.gMASTERHEX.BRANCHDIST.gHEX[.dirty]].

        For example, 'v1.2.3+4.g1abcdef.5.g2345678'
        meaning 4 commits from v1.2.3 to master, and 5 commits from master to the current branch.

        Exceptions:
            1. no tags. git_describe was just HEX. 0+untagged.MASTERDISTANCE.gMASTERHEX.BRANCHDIST.gHEX[.dirty]
            2. current branch is master. Just like PEP440 style.
            3. if no master is found after the tag, we assume tag = master. (TAG+0.gTAGHEX.BRANCHDIST.gHEX[.dirty])
                - the logic is in GitMasterDistance.from_git()

        Note:
            - New in Version-Pioneer.
        """
        try:
            master_info = GitMasterDistance.from_git(
                tag_of_interest=self.closest_fulltag, cwd=self.cwd, verbose=self.verbose
            )
        except CurrentBranchIsMasterError:
            # exception #2
            return self._render_pep440()

        if (
            master_info.distance_from_tag_to_master == 0
            and master_info.distance_from_master == 0
        ):
            # Just the tag
            return self.closest_tag or "0+untagged"

        if self.closest_tag:
            rendered = self.closest_tag
            rendered += self._plus_or_dot
        else:
            # exception #1
            rendered = "0+untagged."

        rendered += (
            f"{master_info.distance_from_tag_to_master}.g{master_info.master_commit_short}"
            f".{master_info.distance_from_master}.g{self.short}"
        )

        if self.dirty:
            rendered += ".dirty"

        return rendered

    def _render_pep440_branch(self) -> str:
        """
        TAG[[.dev0]+DISTANCE.gHEX[.dirty]] .

        The ".dev0" means not master branch. Note that .dev0 sorts backwards
        (a feature branch will appear "older" than the master branch).

        Exceptions:
            1: no tags. 0[.dev0]+untagged.DISTANCE.gHEX[.dirty]
        """
        if self.closest_tag:
            rendered = self.closest_tag
            if self.distance or self.dirty:
                if self.branch not in MASTER_BRANCHES:
                    rendered += ".dev0"
                rendered += self._plus_or_dot
                rendered += f"{self.distance}.g{self.short}"
                if self.dirty:
                    rendered += ".dirty"
        else:
            # exception #1
            rendered = "0"
            if self.branch not in MASTER_BRANCHES:
                rendered += ".dev0"
            rendered += f"+untagged.{self.distance}.g{self.short}"
            if self.dirty:
                rendered += ".dirty"
        return rendered

    @staticmethod
    def _pep440_split_post(ver: str) -> "tuple[str, int | None]":
        """
        Split pep440 version string at the post-release segment.

        Returns the release segments before the post-release and the
        post-release version number (or -1 if no post-release segment is present).
        """
        vc = str.split(ver, ".post")
        return vc[0], int(vc[1] or 0) if len(vc) == 2 else None

    def _render_pep440_pre(self) -> str:
        """
        TAG[.postN.devDISTANCE] -- No -dirty.

        Exceptions:
            1: no tags. 0.post0.devDISTANCE
        """
        if self.closest_tag:
            if self.distance:
                # update the post release segment
                tag_version, post_version = self._pep440_split_post(self.closest_tag)
                rendered = tag_version
                if post_version is not None:
                    rendered += f".post{post_version}.dev{self.distance}"
                else:
                    rendered += f".post0.dev{self.distance}"
            else:
                # no commits, use the tag as the version
                rendered = self.closest_tag
        else:
            # exception #1
            rendered = f"0.post0.dev{self.distance}"
        return rendered

    def _render_pep440_post(self) -> str:
        """
        TAG[.postDISTANCE+gHEX[.dirty]] .

        Note:
            Difference from versioneer is that .dev0 used to be used for .dirty. Their note:

            > TAG[.postDISTANCE[.dev0]+gHEX] .
            >
            > The ".dev0" means dirty. Note that .dev0 sorts backwards
            > (a dirty tree will appear "older" than the corresponding clean one),
            > but you shouldn't be releasing software with -dirty anyways.

        Exceptions:
            When no tags: 0.postDISTANCE+gHEX[.dirty]
        """
        if self.closest_tag:
            rendered = self.closest_tag
            if self.distance or self.dirty:
                rendered += f".post{self.distance}"
                rendered += self._plus_or_dot
                rendered += f"g{self.short}"
                if self.dirty:
                    rendered += ".dirty"
        else:
            # exception
            rendered = f"0.post{self.distance}"
            rendered += f"+g{self.short}"
            if self.dirty:
                rendered += ".dirty"
        return rendered

    def _render_pep440_post_branch(self) -> str:
        """
        TAG[.postDISTANCE[.dev0]+gHEX[.dirty]] .

        The ".dev0" means not master branch.

        Exceptions:
        1: no tags. 0.postDISTANCE[.dev0]+gHEX[.dirty]
        """
        if self.closest_tag:
            rendered = self.closest_tag
            if self.distance or self.dirty:
                rendered += f".post{self.distance}"
                if self.branch not in MASTER_BRANCHES:
                    rendered += ".dev0"
                rendered += self._plus_or_dot
                rendered += f"g{self.short}"
                if self.dirty:
                    rendered += ".dirty"
        else:
            # exception #1
            rendered = f"0.post{self.distance}"
            if self.branch not in MASTER_BRANCHES:
                rendered += ".dev0"
            rendered += f"+g{self.short}"
            if self.dirty:
                rendered += ".dirty"
        return rendered

    def _render_git_describe(self) -> str:
        """
        TAG[-DISTANCE-gHEX][-dirty].

        Like 'git describe --tags --dirty --always'.

        Exceptions:
        1: no tags. HEX[-dirty]  (note: no 'g' prefix)
        """
        if self.closest_tag:
            rendered = self.closest_tag
            if self.distance:
                rendered += f"-{self.distance}-g{self.short}"
        else:
            # exception #1
            rendered = self.short
        if self.dirty:
            rendered += "-dirty"
        return rendered

    def _render_git_describe_long(self) -> str:
        """
        TAG-DISTANCE-gHEX[-dirty].

        Like 'git describe --tags --dirty --always -long'.
        The distance/hash is unconditional.

        Exceptions:
        1: no tags. HEX[-dirty]  (note: no 'g' prefix)
        """
        if self.closest_tag:
            rendered = self.closest_tag
            rendered += f"-{self.distance}-g{self.short}"
        else:
            # exception #1
            rendered = self.short
        if self.dirty:
            rendered += "-dirty"
        return rendered

    def _render_digits(self) -> str:
        """
        TAG.DISTANCE.

        For example, 'v1.2.3+4.g1abcdef' -> '1.2.3.4' and
        'v1.2.3+4.g1abcdef.dirty' -> '1.2.3.5' (dirty counts as 1 commit further).

        Digit-only version string that is compatible with most package managers.

        Note:
            - New in Version-Pioneer.
            - Compatible with Chrome extension version format.
                - Chrome extension version should not have more than 4 segments,
                  so make sure the tags are up to 3 segments.
        """
        if self.error:
            raise ValueError("Unable to render version")

        if self.closest_tag:
            closest_tag: str = self.closest_tag
        else:
            closest_tag = "0"

        version = closest_tag
        if self.distance or self.dirty:
            if self.dirty:
                version += f".{self.distance + 1}"
            else:
                version += f".{self.distance}"

        return version

    def render(self, style: VersionStyle = VersionStyle.pep440) -> VersionDict:
        """Render the given version pieces into the requested style."""
        if self.error:
            return {
                "version": "unknown",
                "full_revisionid": self.long,
                "dirty": None,
                "error": self.error,
                "date": None,
            }

        if style == "pep440":
            rendered = self._render_pep440()
        elif style == "pep440-master":
            rendered = self._render_pep440_master()
        elif style == "pep440-branch":
            rendered = self._render_pep440_branch()
        elif style == "pep440-pre":
            rendered = self._render_pep440_pre()
        elif style == "pep440-post":
            rendered = self._render_pep440_post()
        elif style == "pep440-post-branch":
            rendered = self._render_pep440_post_branch()
        elif style == "git-describe":
            rendered = self._render_git_describe()
        elif style == "git-describe-long":
            rendered = self._render_git_describe_long()
        elif style == "digits":
            rendered = self._render_digits()
        else:
            raise ValueError(f"unknown style '{style}'")

        return {
            "version": rendered,
            "full_revisionid": self.long,
            "dirty": self.dirty,
            "error": None,
            "date": self.date,
        }


def _run_command(
    commands: "list[str]",
    args: "list[str | PathLike]",
    *,
    cwd: "str | PathLike | None" = None,
    hide_stderr: bool = False,
    env: "dict[str, str] | None" = None,
    verbose: bool = False,
) -> "tuple[str | None, int | None]":
    """Call the given command(s)."""
    assert isinstance(commands, list)
    process = None

    popen_kwargs: dict[str, Any] = {}
    if sys.platform == "win32":
        # This hides the console window if pythonw.exe is used
        startupinfo = subprocess.STARTUPINFO()
        startupinfo.dwFlags |= subprocess.STARTF_USESHOWWINDOW
        popen_kwargs["startupinfo"] = startupinfo

    for command in commands:
        dispcmd = str([command, *args])
        try:
            # remember shell=False, so use git.cmd on windows, not just git
            process = subprocess.Popen(
                [command, *args],
                cwd=cwd,
                env=env,
                stdout=subprocess.PIPE,
                stderr=(subprocess.PIPE if hide_stderr else None),
                **popen_kwargs,
            )
            break
        except OSError as e:
            if e.errno == errno.ENOENT:
                continue
            if verbose:
                print(f"unable to run {dispcmd}")
                print(e)
            return None, None
    else:
        if verbose:
            print(f"unable to find command, tried {commands}")
        return None, None
    stdout = process.communicate()[0].strip().decode()
    if process.returncode != 0:
        if verbose:
            print(f"unable to run {dispcmd} (error)")
            print(f"stdout was {stdout}")
        return None, process.returncode
    return stdout, process.returncode


def _run_git_command_or_error(
    args: "list[str | PathLike]",
    *,
    cwd: "str | PathLike | None" = None,
    env: "dict[str, str] | None" = None,
    verbose: bool = False,
):
    """Run a git command or raise an error."""
    out, rc = _run_command(
        GIT_COMMANDS, args, cwd=cwd, hide_stderr=not verbose, env=env, verbose=verbose
    )

    if rc != 0 or out is None:
        raise NotThisMethodError(f"'git {' '.join(map(str, args))}' returned error")
    return out, rc


def _find_root_dir_with_file(
    source: "str | PathLike", marker: "str | Iterable[str]"
) -> Path:
    """
    Find the first parent directory containing a specific "marker", relative to a file path.
    """
    source = Path(source).resolve()
    if isinstance(marker, str):
        marker = {marker}

    while source != source.parent:
        if any((source / m).exists() for m in marker):
            return source

        source = source.parent

    raise FileNotFoundError(f"File {marker} not found in any parent directory")


def get_version_from_parentdir(
    parentdir_prefix: "str | None", cwd: "str | PathLike", *, verbose: bool = False
) -> VersionDict:
    """
    Try to determine the version from the parent directory name.

    Source tarballs conventionally unpack into a directory that includes both the project name and a version string.

    If parentdir_prefix is None, it will try to determine the parentdir_prefix from pyproject.toml.
    """
    rootdirs = []

    # First find a directory with `pyproject.toml`, `setup.cfg`, or `setup.py`
    root = _find_root_dir_with_file(cwd, ["pyproject.toml", "setup.cfg", "setup.py"])

    def try_parentdir(
        project_root: Path, parentdir_prefix: str
    ) -> "VersionDict | None":
        # It's likely that the root is the parent directory of the package,
        # but in some cases like multiple languages, mono-repo, etc. it may not be.
        for _ in range(3):
            dirname = project_root.name
            if dirname.startswith(parentdir_prefix):
                return {
                    "version": dirname[len(parentdir_prefix) :],
                    "full_revisionid": None,
                    "dirty": False,
                    "error": None,
                    "date": None,
                }
            rootdirs.append(project_root)

            if project_root.parent.samefile(project_root):
                break
            project_root = project_root.parent

        return None

    def get_prefix_from_source_url(source_url: str) -> "str | None":
        if not source_url.startswith(
            "https://github.com/"
        ) and not source_url.startswith("https://gitlab.com/"):
            return None

        # Remove trailing .git
        if source_url.endswith(".git"):
            source_url = source_url[:-4]

        # Last part of the URL plus a hyphen
        return source_url.split("/")[-1] + "-"

    def try_parentdir_from_source_url(source_url: str) -> "VersionDict | None":
        prefix = get_prefix_from_source_url(source_url)
        if prefix:
            version_dict = try_parentdir(root, prefix)
            if version_dict:
                return version_dict
        return None

    def try_all_parentdir_in_pyproject_toml(pyproject: dict) -> "VersionDict | None":
        with contextlib.suppress(KeyError):
            version_dict = try_parentdir_from_source_url(
                pyproject["project"]["urls"]["homepage"]
            )
            if version_dict:
                return version_dict
        with contextlib.suppress(KeyError):
            version_dict = try_parentdir_from_source_url(
                pyproject["project"]["urls"]["Homepage"]
            )
            if version_dict:
                return version_dict
        with contextlib.suppress(KeyError):
            version_dict = try_parentdir_from_source_url(
                pyproject["project"]["urls"]["source"]
            )
            if version_dict:
                return version_dict
        with contextlib.suppress(KeyError):
            version_dict = try_parentdir_from_source_url(
                pyproject["project"]["urls"]["Source"]
            )
            if version_dict:
                return version_dict
            return None
        with contextlib.suppress(KeyError):
            return try_parentdir(root, pyproject["project"]["name"] + "-")

    if parentdir_prefix is None:
        # NOTE: New in Version-Pioneer
        # Automatically determine the parentdir_prefix from pyproject.toml
        # 1. project.urls.Homepage
        # homepage / Homepage / source / Source -> if https://github.com/ -> remove trailing .git
        """
        # numpy example
        [project.urls]
        homepage = "https://numpy.org"
        documentation = "https://numpy.org/doc/"
        source = "https://github.com/numpy/numpy"
        download = "https://pypi.org/project/numpy/#files"
        tracker = "https://github.com/numpy/numpy/issues"
        "release notes" = "https://numpy.org/doc/stable/release"
        """
        # NOTE: 2. project.name
        """
        [project]
        name = "version-pioneer"
        """

        if sys.version_info >= (3, 11):
            import tomllib
        else:
            try:
                import tomli as tomllib
            except ModuleNotFoundError as e:
                raise ModuleNotFoundError(
                    "tomli not found. Install it with `pip install tomli`"
                ) from e

        with open(root / "pyproject.toml", "rb") as f:
            pyproject = tomllib.load(f)
        version_dict = try_all_parentdir_in_pyproject_toml(pyproject)

    else:
        version_dict = try_parentdir(root, parentdir_prefix)

    if version_dict:
        return version_dict

    if verbose:
        print(
            f"Tried directories {rootdirs!s} but none started with prefix {parentdir_prefix or '<auto>'}"
        )
    raise NotThisMethodError("rootdir doesn't start with parentdir_prefix")


def get_version_from_pkg_info(cwd: "str | PathLike") -> VersionDict:
    """
    Parse PKG-INFO file if it exists, because it's the most reliable way
    to get the version from an sdist (`build --sdist`)
    since sdist would not have a git information.

    This matters if you choose to "NOT" write the versionfile.

    i.e. [tool.version-pioneer]
         versionscript = src/my_package/_version.py
         # versionfile-sdist = NOT DEFINED

    Note:
        - New in Version-Pioneer
        - Hatchling's Version Source Plugin is deactivated when PKG-INFO is present, so this method would not matter.
        - Only for other backends like setuptools.
    """
    try:
        project_root = _find_root_dir_with_file(cwd, "PKG-INFO")
    except FileNotFoundError:
        raise NotThisMethodError("PKG-INFO not found")  # noqa: B904
    else:
        pyproject_toml = project_root / "pyproject.toml"
        if not pyproject_toml.exists():
            raise NotThisMethodError(
                "PKG-INFO found but no pyproject.toml found in the project root"
            )

        # Confirm [tool.version-pioneer] section exists in pyproject.toml
        with open(pyproject_toml) as f:
            lines = f.readlines()
        for line in lines:
            if "[tool.version-pioneer]" in line:
                break
        else:
            raise NotThisMethodError(
                "[tool.version-pioneer] section not found in pyproject.toml"
            )

        # Read PKG-INFO file
        with open(project_root / "PKG-INFO") as f:
            pkg_info = Parser().parse(f)
        pkg_version = pkg_info.get("Version")
        if not pkg_version:
            raise NotThisMethodError("Version not found in PKG-INFO")

        return {
            "version": pkg_version,
            "full_revisionid": None,
            "dirty": False,
            "error": None,
            "date": None,
        }


def get_version_dict_with_all_methods(
    cfg: "VersionPioneerConfig | None" = None, *, cwd: "str | PathLike | None" = None
) -> VersionDict:
    """
    Get version information from PKG-INFO, Git tags or parent directory as a fallback.
    """
    if cfg is None:
        cfg = VersionPioneerConfig()

    if cwd is None:
        cwd = _SCRIPT_DIR_OR_CURRENT_DIR

    try:
        return get_version_from_pkg_info(cwd)
    except NotThisMethodError:
        pass

    try:
        return GitPieces.from_git(cfg.tag_prefix, cwd=cwd, verbose=cfg.verbose).render(
            cfg.style
        )
    except NotThisMethodError:
        pass

    try:
        return get_version_from_parentdir(
            cfg.parentdir_prefix, cwd, verbose=cfg.verbose
        )
    except NotThisMethodError:
        pass

    return {
        "version": "0+unknown",
        "full_revisionid": None,
        "dirty": None,
        "error": "unable to compute version",
        "date": None,
    }


# IMPORTANT: However you customise the file, make sure the following function is defined!
def get_version_dict() -> VersionDict:
    return get_version_dict_with_all_methods()


if __name__ == "__main__":
    import json

    print(json.dumps(get_version_dict()))



from __future__ import annotations

import contextlib
import subprocess
import sys
from io import StringIO
from os import PathLike
from pathlib import Path


def _run_module(*args, check=True):
    """
    Run python module like `python -m ...`.
    """
    process = subprocess.run(
        [sys.executable, "-m", *args],
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        encoding="utf-8",
        check=check,
    )

    return process


def build_project(*args, cwd=None, check=True, use_uv=True):
    """
    By default, build both wheel and sdist. And just check the content of the wheel later.

    If the wheel is built correctly this way, the sdist should be correct as well. (project dir -> sdist -> wheel)
    But if you build them separately, the sdist is skipped so we can't be sure.
    """
    if not use_uv:
        # replace --out-dir with --outdir (pyproject-build uses --outdir)
        args = (arg if arg != "--out-dir" else "--outdir" for arg in args)
        process = _run_module("build", *args, check=check)
        output = process.stdout

    else:
        process = subprocess.run(
            ["uv", "build", *args],
            check=check,
            capture_output=True,
            text=True,
            encoding="utf-8",
            cwd=cwd,
        )
        output = process.stderr

    # Find Successfully built *.whl
    built_paths: list[Path] = []
    for line in output.splitlines():
        if line.startswith("Successfully built"):
            # if not wheel_path.is_absolute():
            #     # pyproject-build does not print absolute path
            #     wheel_path = Path(d) / wheel_path
            built_paths.append(Path(line.split()[2]))

    # if not built_paths:
    #     raise RuntimeError("Failed to build plugin")

    return process.stderr, built_paths


def unpack_wheel(wheel_path: str | PathLike, dest_dir: str | PathLike | None = None):
    from wheel.cli.unpack import unpack

    f = StringIO()
    with contextlib.redirect_stdout(f):
        unpack(str(wheel_path), str(dest_dir) if dest_dir else ".")
    return f.getvalue()



from __future__ import annotations


def unidiff_output(expected: str, actual: str):
    """Returns a string containing the unified diff of two multiline strings."""
    import difflib

    expected_list = expected.splitlines(keepends=True)
    actual_list = actual.splitlines(keepends=True)

    diff = difflib.unified_diff(expected_list, actual_list)

    return "".join(diff)



# ruff: noqa: T201
from __future__ import annotations

import stat
from pathlib import Path

from pdm.backend.hooks.base import Context

from version_pioneer.utils.toml import get_toml_value
from version_pioneer.utils.version_script import (
    convert_version_dict,
    exec_version_script,
    find_version_script_from_pyproject_toml_dict,
)


class VersionPioneerBuildHook:
    def pdm_build_initialize(self, context: Context):
        # Update metadata version
        versionscript = find_version_script_from_pyproject_toml_dict(
            context.config.data, either_versionfile_or_versionscript=True
        )
        version_dict = exec_version_script(versionscript)
        context.config.metadata["version"] = version_dict["version"]

        # Write the static version file
        if context.target != "editable":
            try:
                if context.target == "wheel":
                    versionfile = context.build_dir / Path(
                        get_toml_value(
                            context.config.data,
                            ["tool", "version-pioneer", "versionfile-wheel"],
                            raise_error=True,
                        )
                    )
                elif context.target == "sdist":
                    versionfile = context.build_dir / Path(
                        get_toml_value(
                            context.config.data,
                            ["tool", "version-pioneer", "versionfile-sdist"],
                            raise_error=True,
                        )
                    )
                else:
                    raise ValueError(f"Unsupported target: {context.target}")
            except KeyError as e:
                print(str(e))  # Missing versionfile-sdist/build in pyproject.toml
                print("Skipping writing a constant version file")
            else:
                context.ensure_build_dir()
                versionfile.parent.mkdir(parents=True, exist_ok=True)
                versionfile.write_text(
                    convert_version_dict(version_dict, output_format="python")
                )
                # make it executable
                versionfile.chmod(versionfile.stat().st_mode | stat.S_IEXEC)



from __future__ import annotations

import subprocess
import tempfile
import textwrap
from collections.abc import Sequence
from os import PathLike
from pathlib import Path
from shutil import rmtree
from typing import Literal, overload

from version_pioneer.utils.files import (
    are_dir_trees_equal,
    find_root_dir_with_file,
    remove_files_recusively,
)
from version_pioneer.utils.version_script import (
    RESOLUTION_FORMAT_TYPE,
    ResolutionFormat,
    convert_version_dict,
)
from version_pioneer.versionscript import (
    VERSION_STYLE_TYPE,
    VersionPioneerConfig,
    VersionStyle,
    get_version_dict_with_all_methods,
)


class VersionMismatchError(Exception):
    pass


def get_version_script_core_code():
    """Get the content of versionscript.py file."""
    from version_pioneer import __version__
    from version_pioneer.versionscript import __file__ as VERSIONSCRIPT_FILE

    version_py_code = Path(VERSIONSCRIPT_FILE).read_text()

    # Put header after the shebang line
    version_py_code = version_py_code.replace(
        "#!/usr/bin/env python3",
        textwrap.dedent(f"""
            #!/usr/bin/env python3
            # THIS "versionscript" IS GENERATED BY version-pioneer-{__version__}
        """).strip(),
        1,
    )

    return version_py_code


def exec_version_script_and_convert(
    project_dir_or_version_script_file: str | PathLike | None = None,
    *,
    output_format: RESOLUTION_FORMAT_TYPE = ResolutionFormat.version_string,
):
    """
    Resolve the _version.py file for build, and return in python, string or json format.

    Examples:
        $ version-pioneer resolve-version --output-format=json
        {"version": "1.2.3", "full_revisionid": "xxxxxx", "dirty": False, "error": None, "date": "2024-12-17T12:25:42+0900"}

        $ version-pioneer resolve-version --output-format=python
        #!/usr/bin/env python3
        # GENERATED BY version-pioneer-v0.1.0
        def get_version_dict():
            return {"version": "0.3.2+15.g2127fd3.dirty", "full-revisionid": "2127fd373d14ed5ded497fc18ac1c1b667f93a7d", "dirty": True, "error": None, "date": "2024-12-17T12:25:42+0900"}

        if __name__ == "__main__":
            import json

            print(json.dumps(get_version_dict()))
    """  # noqa: E501
    from version_pioneer.utils.version_script import (
        exec_version_script,
        find_version_script_from_project_dir,
    )

    if project_dir_or_version_script_file is None:
        version_py_file = find_version_script_from_project_dir(
            Path.cwd(), either_versionfile_or_versionscript=True
        )
    else:
        project_dir_or_version_script_file = Path(project_dir_or_version_script_file)
        if project_dir_or_version_script_file.is_file():
            version_py_file = project_dir_or_version_script_file
        else:
            version_py_file = find_version_script_from_project_dir(
                project_dir_or_version_script_file,
                either_versionfile_or_versionscript=True,
            )

    version_dict = exec_version_script(version_py_file)
    return convert_version_dict(version_dict, output_format)


def get_version_dict_wo_exec(
    cwd: str | PathLike | None = None,
    *,
    style: VERSION_STYLE_TYPE = VersionStyle.pep440,
    tag_prefix: str = "v",
    parentdir_prefix: str | None = None,
    verbose: bool = False,
):
    """
    WITHOUT using the installed _version.py file, get version with Version-Pioneer logic.

    Useful when you don't need to customise the _version.py file, and you work in non-Python projects
    so you don't care about re-evaluating the version file.

    Args:
        parentdir_prefix: The prefix of the parent directory. (e.g. {github-repo-name}-)
            None refers auto-detection from pyproject.toml.
    """
    cfg = VersionPioneerConfig(
        style=VersionStyle(style),
        tag_prefix=tag_prefix,
        parentdir_prefix=parentdir_prefix,
        verbose=verbose,
    )

    version_dict = get_version_dict_with_all_methods(
        cfg, cwd=Path.cwd() if cwd is None else cwd
    )
    return version_dict


def get_version_wo_exec_and_convert(
    cwd: str | PathLike | None = None,
    *,
    style: VERSION_STYLE_TYPE = VersionStyle.pep440,
    tag_prefix: str = "v",
    parentdir_prefix: str | None = None,
    output_format: RESOLUTION_FORMAT_TYPE = ResolutionFormat.version_string,
    verbose: bool = False,
):
    """
    WITHOUT using the installed _version.py file, get version with Version-Pioneer logic, and return as a string.

    Useful when you don't need to customise the _version.py file, and you work in non-Python projects
    so you don't care about re-evaluating the version file.

    Args:
        project_dir: The root or child directory of the project.
        parentdir_prefix: The prefix of the parent directory. (e.g. {github_repo_name}-)
    """
    return convert_version_dict(
        get_version_dict_wo_exec(
            cwd=cwd,
            style=style,
            tag_prefix=tag_prefix,
            parentdir_prefix=parentdir_prefix,
            verbose=verbose,
        ),
        output_format,
    )


def _get_wheel_package_version(wheel_path: str | PathLike) -> str:
    # also works with sdist
    return Path(wheel_path).stem.split(".tar")[0].split("-")[1]


def _get_wheel_package_name_and_version(wheel_path: str | PathLike) -> str:
    split = Path(wheel_path).stem.split("-")
    return "-".join(split[:2])


# NOTE: if you name it with `test_`, it will be picked up by pytest by importing this module.
@overload
def build_consistency_test(
    project_dir: str | PathLike | None = None,
    *,
    delete_temp_dir: Literal[True],
    test_chaining: bool = True,
    expected_version: str | None = None,
    ignore_patterns: str | Sequence[str] = ("*.egg-info/SOURCES.txt"),
) -> None: ...


@overload
def build_consistency_test(
    project_dir: str | PathLike | None = None,
    *,
    delete_temp_dir: Literal[False],
    test_chaining: bool = True,
    expected_version: str | None = None,
    ignore_patterns: str | Sequence[str] = ("*.egg-info/SOURCES.txt"),
) -> Path: ...


@overload
def build_consistency_test(
    project_dir: str | PathLike | None = None,
    *,
    delete_temp_dir: bool = True,
    test_chaining: bool = True,
    expected_version: str | None = None,
    ignore_patterns: str | Sequence[str] = ("*.egg-info/SOURCES.txt"),
) -> Path | None: ...


def build_consistency_test(
    project_dir: str | PathLike | None = None,
    *,
    delete_temp_dir: bool = True,
    test_chaining: bool = True,
    expected_version: str | None = None,
    ignore_patterns: str | Sequence[str] = ("*.egg-info/SOURCES.txt"),
) -> Path | None:
    """
    Assure build equality with `build`, `build --wheel` and `build --sdist`,
    plus check chaining sdist builds (project -> sdist -> sdist).

    This is because (I assume) when building both wheel and sdist, it first generates sdist (which should include all
    files necessary for the wheel), then builds the wheel from the sdist. If you build with `build --wheel` only, it
    processes the files from the git project directory directly. Thus, it is likely that `build` (both) fails to include
    version from git describe, while `build --wheel` does.

    Also, the _version.py gets resolved twice, once for the sdist, and once for the wheel. This can result in different
    outputs, thus we need to check the equality.

    Note:
        - setuptools backend generates setup.cfg in the sdist, so *.egg-info/SOURCES.txt can be different.
        - Other backends may generate different files, so you may need to add more ignore patterns.
    """
    import verboselogs

    from version_pioneer.utils.build import build_project, unpack_wheel

    def _compare_tmp_dirs(dir1: Path, dir2: Path, *, error_msg: str):
        remove_files_recusively(dir1, patterns=ignore_patterns)
        remove_files_recusively(dir2, patterns=ignore_patterns)
        try:
            are_dir_trees_equal(
                dir1,
                dir2,
            )
        except FileNotFoundError as e:
            raise FileNotFoundError(error_msg) from e

    logger = verboselogs.VerboseLogger(__name__)

    ignore_patterns = (
        [ignore_patterns] if isinstance(ignore_patterns, str) else ignore_patterns
    )

    # Ensure command `uv` is available
    try:
        subprocess.run(["uv", "--version"], check=True)
    except FileNotFoundError:
        logger.error(
            "❌ Command `uv` is not available. Please install with `pip install uv` or `brew install uv`."
        )
        return None

    logger.info("Testing build consistency...")

    project_dir = Path.cwd() if project_dir is None else Path(project_dir)
    project_root = find_root_dir_with_file(project_dir, "pyproject.toml")

    # We don't need to change the cwd, as the build_project function does that.
    logger.info(f"Changing cwd to {project_root}")
    # os.chdir(project_root)

    # build the project
    temp_dir = Path(tempfile.mkdtemp())
    logger.info("Building the project with `uv build`")
    output, builds = build_project(
        "--out-dir", temp_dir / "dist-combined", cwd=project_root
    )

    assert len(builds) == 2, f"❌ Expected 2 builds. {output}"
    sdist_combined = builds[0]
    wheel_combined = builds[1]

    logger.info("Building the project with `uv build --sdist` and `uv build --wheel`")
    output, builds = build_project(
        "--sdist", "--out-dir", temp_dir / "dist", cwd=project_root
    )
    assert len(builds) == 1, f"❌ Expected 1 sdist to be built. {output}"
    sdist_separate = builds[0]

    output, builds = build_project(
        "--wheel", "--out-dir", temp_dir / "dist", cwd=project_root
    )
    assert len(builds) == 1, f"❌ Expected 1 wheel to be built. {output}"
    wheel_separate = builds[0]

    wheel_combined_version = _get_wheel_package_version(wheel_combined)
    wheel_separate_version = _get_wheel_package_version(wheel_separate)
    sdist_combined_version = _get_wheel_package_version(sdist_combined)
    sdist_separate_version = _get_wheel_package_version(sdist_separate)
    if expected_version is not None:
        assert (
            wheel_combined_version == expected_version
        ), f"❌ Expected version {expected_version}, but got {wheel_combined_version}"
    if not (
        wheel_combined_version
        == wheel_separate_version
        == sdist_combined_version
        == sdist_separate_version
    ):
        raise VersionMismatchError(
            f"❌ Versions are not consistent. {wheel_combined_version=}, {wheel_separate_version=}, "
            f"{sdist_combined_version=}, {sdist_separate_version=}",
            temp_dir,
        )

    if expected_version is None:
        expected_version = wheel_combined_version
        logger.info(f"Built version: {expected_version}")
    else:
        logger.info(f"✅ Built version: {expected_version}")

    unpack_wheel(wheel_combined, temp_dir / "dist-combined")
    unpack_wheel(wheel_separate, temp_dir / "dist")
    dir_name = _get_wheel_package_name_and_version(wheel_combined)

    _compare_tmp_dirs(
        temp_dir / "dist" / dir_name,
        temp_dir / "dist-combined" / dir_name,
        error_msg="❌ Wheel builds are not consistent.",
    )

    rmtree(temp_dir / "dist-combined" / dir_name)
    rmtree(temp_dir / "dist" / dir_name)

    logger.success("✅ 2 wheel builds are consistent.")

    subprocess.run(
        ["tar", "-xzf", sdist_separate, "--directory", temp_dir / "dist"], check=True
    )
    subprocess.run(
        ["tar", "-xzf", sdist_separate, "--directory", temp_dir / "dist-combined"],
        check=True,
    )

    _compare_tmp_dirs(
        temp_dir / "dist" / dir_name,
        temp_dir / "dist-combined" / dir_name,
        error_msg="❌ sdist builds are not consistent.",
    )

    rmtree(temp_dir / "dist-combined")
    # rmtree(temp_dir / "dist" / dir_name)

    logger.success("✅ 2 sdist builds are consistent.")

    if test_chaining:
        logger.info(
            "Building the project with `uv build --sdist` using the built sdist (chaining test)."
        )
        built_dir = temp_dir / "dist" / dir_name
        logger.info(f"Changing cwd to the built sdist directory: {built_dir}")
        # os.chdir(built_dir)

        # Ensure no git information is available
        ps = subprocess.run(
            ["git", "describe"],
            cwd=built_dir,
            check=False,
            capture_output=True,
            text=True,
        )
        assert (
            ps.returncode == 128
        ), "❌ Your temp directory has a .git information. Not able to test properly..."

        output, builds = build_project(
            "--sdist", "--out-dir", temp_dir / "dist-chained", cwd=built_dir
        )
        assert len(builds) == 1, f"❌ Expected 1 sdist to be built. {output}"
        sdist_chained = builds[0]

        sdist_chained_version = _get_wheel_package_version(sdist_chained)
        if expected_version != sdist_chained_version:
            raise VersionMismatchError(
                f"❌ Versions are not consistent. {expected_version=}, {sdist_chained_version=}",
                temp_dir,
            )

        subprocess.run(
            ["tar", "-xzf", sdist_chained, "--directory", temp_dir / "dist-chained"],
            check=True,
        )
        # building may generate a build artifact like .pdm_build, so we reset the `built_dir`.
        rmtree(built_dir)
        subprocess.run(
            ["tar", "-xzf", sdist_separate, "--directory", temp_dir / "dist"],
            check=True,
        )
        built_dir = temp_dir / "dist" / dir_name

        _compare_tmp_dirs(
            built_dir,
            temp_dir / "dist-chained" / dir_name,
            error_msg="❌ Chained sdist builds are not consistent with the non-chained build."
            f"original={built_dir}  "
            f"chained={temp_dir / 'dist-chained' / dir_name}",
        )

        logger.success("✅ Chained sdist builds are consistent.")
        rmtree(temp_dir / "dist-chained")

        logger.info("Build wheel using the sdist.")
        output, builds = build_project(
            "--wheel", "--out-dir", temp_dir / "dist-chained", cwd=built_dir
        )
        assert len(builds) == 1, f"❌ Expected 1 wheel to be built. {output}"
        wheel_chained = builds[0]

        wheel_chained_version = _get_wheel_package_version(wheel_chained)
        if expected_version != wheel_chained_version:
            raise VersionMismatchError(
                f"❌ Versions are not consistent. {expected_version=}, {wheel_chained_version=}",
                temp_dir,
            )

        rmtree(built_dir)
        unpack_wheel(wheel_separate, temp_dir / "dist")
        unpack_wheel(wheel_chained, temp_dir / "dist-chained")

        _compare_tmp_dirs(
            built_dir,
            temp_dir / "dist-chained" / dir_name,
            error_msg="❌ Chained wheel build is not consistent with the non-chained build."
            f"original={built_dir}  "
            f"chained={temp_dir / 'dist-chained' / dir_name}",
        )

        logger.success(
            "✅ sdist -> wheel chained build is consistent with the non-chained build."
        )

        logger.success(
            "💓 All tests passed! 3 sdist builds and 3 wheel builds are consistent."
        )
    else:
        logger.success(
            "💓 All tests passed! 2 sdist builds and 2 wheel builds are consistent."
        )

    if delete_temp_dir:
        logger.info(f"Deleting temporary directory {temp_dir}")
        rmtree(temp_dir)
        return None
    else:
        logger.info(f"Temporary directory is at {temp_dir}")
        return temp_dir



from hatchling.version.source.plugin.interface import VersionSourceInterface

from version_pioneer.utils.version_script import (
    exec_version_script,
    find_version_script_from_project_dir,
)


class VersionPioneerVersionSource(VersionSourceInterface):
    PLUGIN_NAME = "version-pioneer"

    def get_version_data(self):
        versionscript = find_version_script_from_project_dir(
            project_dir=self.root,
            either_versionfile_or_versionscript=True,
        )
        version_dict = exec_version_script(versionscript)

        return {"version": version_dict["version"]}



#!/usr/bin/env python3
# THIS "versionfile" IS GENERATED BY version-pioneer-0.0.8
# by evaluating the original versionscript and storing the computed versions as a constant.

def get_version_dict():
    return {'version': '0.0.8', 'full_revisionid': 'a68da113234a5eb656308871480254f343986aa0', 'dirty': False, 'error': None, 'date': '2024-12-26T15:57:26+0000'}


if __name__ == "__main__":
    import json

    print(json.dumps(get_version_dict()))


# https://github.com/fastapi/typer/issues/336#issuecomment-2434726193
# Add typer help text from docstring
import inspect
from collections.abc import Callable
from functools import wraps
from typing import Any

import docstring_parser
import typer
from typing_extensions import Annotated, get_args, get_origin


def from_docstring(command: Callable) -> Callable:
    """
    A decorator that applies help texts from the function's docstring to Typer arguments/options.

    It will only apply the help text if no help text has been explicitly set in the Typer argument/option.

    Args:
        command (Callable): The function to decorate.

    Returns:
        Callable: The decorated function with help texts applied, without overwriting existing settings.
    """
    if command.__doc__ is None:
        return command

    # Parse the docstring and extract parameter descriptions
    docstring = docstring_parser.parse(command.__doc__)
    param_help = {param.arg_name: param.description for param in docstring.params}

    # The commands's full help text (summary + long description)
    command_help = (
        f"{docstring.short_description or ''}\n\n{docstring.long_description or ''}"
    )

    # Get the signature of the original function
    sig = inspect.signature(command)
    parameters = sig.parameters

    @wraps(command)
    def wrapper(**kwargs: Any) -> Any:
        return command(**kwargs)

    # Prepare a new mapping for parameters
    new_parameters = []

    for name, param in parameters.items():
        help_text = param_help.get(
            name, ""
        )  # Get help text from docstring if available

        param_type = (
            param.annotation if param.annotation is not inspect.Parameter.empty else str
        )  # Default to str if no annotation

        # Handle Annotated (e.g., Annotated[int, typer.Argument()] or Annotated[str, typer.Option()])
        # Check if the parameter uses Annotated
        if get_origin(param_type) is Annotated:
            param_type, *metadata = get_args(param_type)
            # Iterate through the metadata to find Typer's Argument or Option
            new_metadata = []
            for m in metadata:
                if isinstance(m, (typer.models.ArgumentInfo, typer.models.OptionInfo)):  # noqa: SIM102
                    # Only add help text if it's not already set
                    if not m.help:
                        m.help = help_text
                new_metadata.append(m)

            # Rebuild the annotated type with updated metadata (python 3.11)
            # new_param = param.replace(annotation=Annotated[param_type, *new_metadata])
            # for python 3.8, this is not perfect ...
            new_param = param.replace(annotation=Annotated[param_type, new_metadata[0]])

        # If it's an Option or Argument directly (e.g., a: int = typer.Option(...))
        elif isinstance(
            param.default, (typer.models.ArgumentInfo, typer.models.OptionInfo)
        ):
            if not param.default.help:
                param.default.help = help_text
            new_param = param

        else:  # noqa: PLR5501
            # If the parameter has no default, treat it as an Argument
            if param.default is inspect.Parameter.empty:
                new_param = param.replace(
                    default=typer.Argument(..., help=help_text), annotation=param_type
                )
            else:
                # If the parameter has a default, treat it as an Option
                new_param = param.replace(
                    default=typer.Option(param.default, help=help_text),
                    annotation=param_type,
                )

        new_parameters.append(new_param)

    # Create a new signature with updated parameters
    new_sig = sig.replace(parameters=new_parameters)

    # Apply the new signature to the wrapper function
    wrapper.__signature__ = new_sig

    wrapper.__doc__ = command_help.strip()

    return wrapper



# ruff: noqa: T201
from __future__ import annotations

import os
import sys
from collections.abc import Callable
from pathlib import Path
from typing import Any

from version_pioneer.api import exec_version_script_and_convert
from version_pioneer.utils.toml import (
    find_pyproject_toml,
    get_toml_value,
    load_toml,
)
from version_pioneer.utils.version_script import (
    convert_version_dict,
    exec_version_script,
    find_version_script_from_project_dir,
    find_version_script_from_pyproject_toml_dict,
)


def get_cmdclass(cmdclass: dict[str, Any] | None = None):
    """
    Get the custom setuptools subclasses used by Versioneer.

    If the package uses a different cmdclass (e.g. one from numpy), it
    should be provide as an argument.
    """
    if "versioneer" in sys.modules:
        del sys.modules["versioneer"]
        # this fixes the "python setup.py develop" case (also 'install' and
        # 'easy_install .'), in which subdependencies of the main project are
        # built (using setup.py bdist_egg) in the same python process. Assume
        # a main project A and a dependency B, which use different versions
        # of Versioneer. A's setup.py imports A's Versioneer, leaving it in
        # sys.modules by the time B's setup.py is executed, causing B to run
        # with the wrong versioneer. Setuptools wraps the sub-dep builds in a
        # sandbox that restores sys.modules to it's pre-build state, so the
        # parent is protected against the child's "import versioneer". By
        # removing ourselves from sys.modules here, before the child build
        # happens, we protect the child from the parent's versioneer too.
        # Also see https://github.com/python-versioneer/python-versioneer/issues/52

    cmds = {} if cmdclass is None else cmdclass.copy()

    # we add "version" to setuptools
    from setuptools import Command

    class CmdVersion(Command):
        description = "report generated version string"
        user_options: list[tuple[str, str, str]] = []  # noqa: RUF012
        boolean_options: list[str] = []  # noqa: RUF012

        def initialize_options(self) -> None:
            pass

        def finalize_options(self) -> None:
            pass

        def run(self) -> None:
            vers = exec_version_script(
                find_version_script_from_project_dir(
                    either_versionfile_or_versionscript=True
                )
            )
            print(f"Version: {vers['version']}")
            print(f" full-revisionid: {vers['full_revisionid']}")
            print(f" dirty: {vers['dirty']}")
            print(f" date: {vers['date']}")
            if vers["error"]:
                print(f" error: {vers['error']}")

    cmds["version"] = CmdVersion

    # we override "build_py" in setuptools
    #
    # most invocation pathways end up running build_py:
    #  distutils/build -> build_py
    #  distutils/install -> distutils/build ->..
    #  setuptools/bdist_wheel -> distutils/install ->..
    #  setuptools/bdist_egg -> distutils/install_lib -> build_py
    #  setuptools/install -> bdist_egg ->..
    #  setuptools/develop -> ?
    #  pip install:
    #   copies source tree to a tempdir before running egg_info/etc
    #   if .git isn't copied too, 'git describe' will fail
    #   then does setup.py bdist_wheel, or sometimes setup.py install
    #  setup.py egg_info -> ?

    # pip install -e . and setuptool/editable_wheel will invoke build_py
    # but the build_py command is not expected to copy any files.

    # we override different "build_py" commands for both environments
    if "build_py" in cmds:
        _build_py: Any = cmds["build_py"]
    else:
        from setuptools.command.build_py import build_py as _build_py

    class CmdBuildPy(_build_py):
        def run(self) -> None:
            _build_py.run(self)
            if getattr(self, "editable_mode", False):
                # During editable installs `.py` and data files are
                # not copied to build_lib
                return
            # now locate _version.py in the new build/ directory and replace
            # it with an updated value
            pyproject_toml_file = find_pyproject_toml()
            pyproject_toml = load_toml(pyproject_toml_file)
            versionfile_wheel: str | None = get_toml_value(
                pyproject_toml,
                ["tool", "version-pioneer", "versionfile-wheel"],
            )
            if versionfile_wheel is not None:
                versionscript = find_version_script_from_pyproject_toml_dict(
                    pyproject_toml, either_versionfile_or_versionscript=True
                )
                target_versionfile_content = exec_version_script_and_convert(
                    versionscript, output_format="python"
                )
                target_versionfile = Path(self.build_lib) / versionfile_wheel
                print(f"UPDATING {target_versionfile}")
                target_versionfile.write_text(target_versionfile_content)

    cmds["build_py"] = CmdBuildPy

    if "build_ext" in cmds:
        _build_ext: Any = cmds["build_ext"]
    else:
        from setuptools.command.build_ext import build_ext as _build_ext

    class CmdBuildExt(_build_ext):
        def run(self) -> None:
            _build_ext.run(self)
            if self.inplace:
                # build_ext --inplace will only build extensions in
                # build/lib<..> dir with no _version.py to write to.
                # As in place builds will already have a _version.py
                # in the module dir, we do not need to write one.
                return
            # now locate _version.py in the new build/ directory and replace
            # it with an updated value
            pyproject_toml_file = find_pyproject_toml()
            pyproject_toml = load_toml(pyproject_toml_file)
            versionfile_wheel: str | None = get_toml_value(
                pyproject_toml,
                ["tool", "version-pioneer", "versionfile-wheel"],
            )
            if versionfile_wheel is not None:
                versionscript = find_version_script_from_pyproject_toml_dict(
                    pyproject_toml, either_versionfile_or_versionscript=True
                )
                target_versionfile_content = exec_version_script_and_convert(
                    versionscript, output_format="python"
                )
                target_versionfile = Path(self.build_lib) / versionfile_wheel
                if not target_versionfile.exists():
                    print(
                        f"Warning: {target_versionfile} does not exist, skipping "
                        "version update. This can happen if you are running build_ext "
                        "without first running build_py."
                    )
                    return
                print(f"UPDATING {target_versionfile}")
                target_versionfile.write_text(target_versionfile_content)

    cmds["build_ext"] = CmdBuildExt

    def _run_directly_inside_source_tree(run_func: Callable):
        pyproject_toml_file = find_pyproject_toml()
        pyproject_toml = load_toml(pyproject_toml_file)
        versionscript: Path | None = get_toml_value(
            pyproject_toml,
            ["tool", "version-pioneer", "versionscript"],
            return_path_object=True,
        )
        if versionscript is None:
            raise ValueError("versionscript is not set in pyproject.toml")
        versionfile_sdist: Path | None = get_toml_value(
            pyproject_toml,
            ["tool", "version-pioneer", "versionfile-sdist"],
            return_path_object=True,
        )
        if versionfile_sdist is None:
            print("Skipping version update due to versionfile-sdist not set.")
            run_func()
            return

        versionscript = pyproject_toml_file.parent / versionscript
        versionfile_sdist = pyproject_toml_file.parent / versionfile_sdist

        if versionscript == versionfile_sdist:
            # HACK: replace _version.py directly in the source tree during build, and restore it.
            target_versionfile = versionscript
            print(f"UPDATING {target_versionfile}")
            target_versionfile_content = exec_version_script_and_convert(
                versionscript, output_format="python"
            )
            original_versionscript_content = versionscript.read_text()
            target_versionfile.write_text(target_versionfile_content)

            run_func()

            target_versionfile.write_text(original_versionscript_content)
        else:
            # HACK: write _version.py directly in the source tree during build.
            target_versionfile = versionfile_sdist
            target_versionfile_content = exec_version_script_and_convert(
                versionscript, output_format="python"
            )
            target_versionfile.write_text(target_versionfile_content)

            run_func()
            # We do not remove the versionfile-sdist. Put it as .gitignore.

    if "cx_Freeze" in sys.modules:  # cx_freeze enabled?
        try:
            from cx_Freeze.command.build_exe import (  # type: ignore
                BuildEXE as _build_exe,  # noqa: N813
            )
        except ImportError:  # cx_Freeze < 6.11
            from cx_Freeze.dist import build_exe as _build_exe  # type: ignore
        # nczeczulin reports that py2exe won't like the pep440-style string
        # as FILEVERSION, but it can be used for PRODUCTVERSION, e.g.
        # setup(console=[{
        #   "version": versioneer.get_version().split("+", 1)[0], # FILEVERSION
        #   "product_version": versioneer.get_version(),
        #   ...

        class CmdBuildEXE(_build_exe):
            def run(self) -> None:
                _run_directly_inside_source_tree(lambda: _build_exe.run(self))

        cmds["build_exe"] = CmdBuildEXE
        del cmds["build_py"]

    if "py2exe" in sys.modules:  # py2exe enabled?
        try:
            from py2exe.setuptools_buildexe import py2exe as _py2exe  # type: ignore
        except ImportError:
            from py2exe.distutils_buildexe import py2exe as _py2exe  # type: ignore

        class CmdPy2EXE(_py2exe):
            def run(self) -> None:
                _run_directly_inside_source_tree(lambda: _py2exe.run(self))

        cmds["py2exe"] = CmdPy2EXE

    # sdist farms its file list building out to egg_info
    if "egg_info" in cmds:
        _egg_info: Any = cmds["egg_info"]
    else:
        from setuptools.command.egg_info import egg_info as _egg_info

    class CmdEggInfo(_egg_info):
        def find_sources(self) -> None:
            # egg_info.find_sources builds the manifest list and writes it
            # in one shot
            super().find_sources()

            # Modify the filelist and normalize it
            # self.filelist.append("versioneer.py")

            pyproject_toml_file = find_pyproject_toml()
            pyproject_toml = load_toml(pyproject_toml_file)
            versionscript = find_version_script_from_pyproject_toml_dict(
                pyproject_toml, either_versionfile_or_versionscript=True
            )

            # There are rare cases where versionscript might not be
            # included by default, so we must be explicit
            self.filelist.append(str(versionscript))

            self.filelist.sort()
            self.filelist.remove_duplicates()

            # The write method is hidden in the manifest_maker instance that
            # generated the filelist and was thrown away
            # We will instead replicate their final normalization (to unicode,
            # and POSIX-style paths)
            from setuptools import unicode_utils

            normalized = [
                unicode_utils.filesys_decode(f).replace(os.sep, "/")
                for f in self.filelist.files
            ]

            manifest_filename = Path(self.egg_info) / "SOURCES.txt"
            manifest_filename.write_text("\n".join(normalized))

    cmds["egg_info"] = CmdEggInfo

    # we override different "sdist" commands for both environments
    if "sdist" in cmds:
        _sdist: Any = cmds["sdist"]
    else:
        from setuptools.command.sdist import sdist as _sdist

    class CmdSdist(_sdist):
        def run(self) -> None:
            pyproject_toml_file = find_pyproject_toml()
            pyproject_toml = load_toml(pyproject_toml_file)
            versionscript = find_version_script_from_pyproject_toml_dict(
                pyproject_toml, either_versionfile_or_versionscript=True
            )
            self.version_dict = exec_version_script(
                pyproject_toml_file.parent / versionscript
            )

            self.versionfile_sdist: Path | None = get_toml_value(
                pyproject_toml,
                ["tool", "version-pioneer", "versionfile-sdist"],
                return_path_object=True,
            )

            # unless we update this, the command will keep using the old
            # version
            self.distribution.metadata.version = self.version_dict["version"]
            return _sdist.run(self)

        def make_release_tree(self, base_dir: str, files: list[str]) -> None:
            _sdist.make_release_tree(self, base_dir, files)
            # now locate _version.py in the new base_dir directory
            # (remembering that it may be a hardlink) and replace it with an
            # updated value

            if self.versionfile_sdist is None:
                print("Skipping version update due to versionfile-sdist not set.")
            else:
                target_versionfile = Path(base_dir) / self.versionfile_sdist
                print(f"UPDATING {target_versionfile}")
                target_versionfile.write_text(
                    convert_version_dict(self.version_dict, output_format="python")
                )

    cmds["sdist"] = CmdSdist

    return cmds



import logging
import subprocess
import textwrap
from pathlib import Path
from shutil import rmtree

import pytest

from tests.build_pipelines import (
    assert_build_and_version_persistence,
)
from tests.utils import (
    VersionScriptResolutionError,
)
from version_pioneer.api import (
    build_consistency_test,
)
from version_pioneer.utils.build import build_project

from .build_pipelines import check_no_versionfile_output

logger = logging.getLogger(__name__)


def test_build_consistency(new_setuptools_project: Path):
    # Reset the project to a known state.
    subprocess.run(["git", "stash", "--all"], cwd=new_setuptools_project, check=True)
    subprocess.run(
        ["git", "checkout", "v0.1.0"], cwd=new_setuptools_project, check=True
    )
    build_consistency_test(project_dir=new_setuptools_project, expected_version="0.1.0")


def test_build_version(new_setuptools_project: Path):
    # Reset the project to a known state.
    subprocess.run(["git", "stash", "--all"], cwd=new_setuptools_project, check=True)
    subprocess.run(
        ["git", "checkout", "v0.1.0"], cwd=new_setuptools_project, check=True
    )
    assert_build_and_version_persistence(new_setuptools_project)


def test_different_versionfile(new_setuptools_project: Path, plugin_wheel: Path):
    # Reset the project to a known state.
    subprocess.run(["git", "stash", "--all"], cwd=new_setuptools_project, check=True)
    subprocess.run(
        ["git", "checkout", "v0.1.0"], cwd=new_setuptools_project, check=True
    )

    pyp = new_setuptools_project / "pyproject.toml"

    pyp.write_text(
        textwrap.dedent(f"""
            [build-system]
            requires = ["setuptools", "version-pioneer @ {plugin_wheel.as_uri()}"]
            build-backend = "setuptools.build_meta"

            [tool.version-pioneer]
            versionscript = "src/my_app/_version.py"
            versionfile-sdist = "src/my_app/_versionfile.py"
            versionfile-wheel = "my_app/_versionfile.py"

            [project]
            name = "my-app"
            dynamic = ["version"]
            requires-python = ">=3.8"
        """),
    )

    subprocess.run(["git", "add", "."], check=True)
    subprocess.run(["git", "commit", "-m", "Second commit"], check=True)
    subprocess.run(["git", "tag", "v0.1.1"], check=True)

    build_consistency_test(project_dir=new_setuptools_project, expected_version="0.1.1")


def test_invalid_config(new_setuptools_project: Path, plugin_wheel: Path):
    """
    Missing config makes the build fail with a meaningful error message.
    """
    pyp = new_setuptools_project / "pyproject.toml"

    # If we leave out the config for good, the plugin doesn't get activated.
    pyp.write_text(
        textwrap.dedent(f"""
            [build-system]
            requires = ["setuptools", "version-pioneer @ {plugin_wheel.as_uri()}"]
            build-backend = "setuptools.build_meta"

            [tool.version-pioneer]
            # versionscript = "src/my_app/_version.py"
            # versionfile-sdist = "src/my_app/_version.py"
            # versionfile-wheel = "my_app/_version.py"

            [project]
            name = "my-app"
            dynamic = ["version"]
        """),
    )

    err, _ = build_project(check=False)

    assert (
        "KeyError: 'Missing key tool.version-pioneer.versionscript in pyproject.toml'"
        in err
    ), err

    pyp.write_text(
        textwrap.dedent(f"""
            [build-system]
            requires = ["setuptools", "version-pioneer @ {plugin_wheel.as_uri()}"]
            build-backend = "setuptools.build_meta"

            [tool.version-pioneer]
            # versionscript = "src/my_app/_version.py"
            versionfile-sdist = "src/my_app/_version.py"
            # versionfile-wheel = "my_app/_version.py"

            [project]
            name = "my-app"
            dynamic = ["version"]
            requires-python = ">=3.8"
        """),
    )

    err, _ = build_project(check=False)

    assert (
        "KeyError: 'Missing key tool.version-pioneer.versionscript in pyproject.toml'"
        in err
    ), err


@pytest.mark.xfail(raises=VersionScriptResolutionError)
def test_no_versionfile_sdist_nor_wheel(
    new_setuptools_project: Path, plugin_wheel: Path
):
    """
    If versionfile-sdist and versionfile-wheel is not configured, the build does NOT FAIL
    but the _version.py file is not updated.
    """
    # Reset the project to a known state.
    subprocess.run(["git", "stash", "--all"], cwd=new_setuptools_project, check=True)
    subprocess.run(
        ["git", "checkout", "v0.1.0"], cwd=new_setuptools_project, check=True
    )

    pyp = new_setuptools_project / "pyproject.toml"

    pyp.write_text(
        textwrap.dedent(f"""
            [build-system]
            requires = ["setuptools", "version-pioneer @ {plugin_wheel.as_uri()}"]
            build-backend = "setuptools.build_meta"

            [tool.version-pioneer]
            versionscript = "src/my_app/_version.py"
            # versionfile-sdist = "src/my_app/_version.py"
            # versionfile-wheel = "my_app/_version.py"

            [project]
            name = "my-app"
            dynamic = ["version"]
            requires-python = ">=3.8"
        """),
    )

    # Can't use dynamic versioning without versionfile-wheel.
    setup_py = new_setuptools_project / "setup.py"
    setup_py.write_text(
        textwrap.dedent("""
            from setuptools import setup
            from version_pioneer.build.setuptools import get_cmdclass

            setup(
                version="0.1.1",
                cmdclass=get_cmdclass(),
            )
        """),
    )

    subprocess.run(["git", "add", "."], check=True)
    subprocess.run(["git", "commit", "-m", "Second commit"], check=True)
    subprocess.run(["git", "tag", "v0.1.1"], check=True)

    # The build should be consistent still, because versionfile-sdist and versionfile-wheel are both not configured.

    # NOTE: since we don't write the versionfile, we normally can't chain the tests.
    # However, hatchling actually reads PKG-INFO metadata in the sdist,
    # and if it's present, the version source is ignored.
    # In setuptools, the `get_version_dict()` includes this logic.
    temp_dir = build_consistency_test(
        project_dir=new_setuptools_project,
        test_chaining=True,
        delete_temp_dir=False,
        expected_version="0.1.1",
    )

    # No need to build again. We check the _version.py file directly on sdist and wheel.
    Path(temp_dir / "dist").rename(new_setuptools_project / "dist")
    rmtree(temp_dir)

    check_no_versionfile_output(cwd=new_setuptools_project)



from __future__ import annotations

from hatchling.builders.hooks.plugin.interface import BuildHookInterface
from hatchling.plugin import hookimpl
from hatchling.version.source.plugin.interface import VersionSourceInterface

from .build_hook import VersionPioneerBuildHook
from .version_source import VersionPioneerVersionSource


@hookimpl
def hatch_register_version_source() -> type[VersionSourceInterface]:
    return VersionPioneerVersionSource


@hookimpl
def hatch_register_build_hook() -> type[BuildHookInterface]:
    return VersionPioneerBuildHook



# ruff: noqa: T201
from __future__ import annotations

import stat
import tempfile
from pathlib import Path
from typing import Any

from hatchling.builders.hooks.plugin.interface import BuildHookInterface

from version_pioneer.utils.toml import get_toml_value, load_toml
from version_pioneer.utils.version_script import (
    convert_version_dict,
    exec_version_script,
    find_version_script_from_project_dir,
)


class VersionPioneerBuildHook(BuildHookInterface):
    PLUGIN_NAME = "version-pioneer"

    def initialize(self, version: str, build_data: dict[str, Any]) -> None:
        """
        Execute either versionscript or versionfile-sdist and write to versionfile-sdist.

        Note:
            versionfile-wheel is not used for hatchling.

        Args:
            version: editable, standard (note there's no separation from sdist and wheel)
        """
        self.temp_version_file = None

        if version == "editable":
            return

        pyproject_toml = load_toml(Path(self.root) / "pyproject.toml")

        # This also checks the valid config, so run it first.
        versionscript = find_version_script_from_project_dir(
            project_dir=self.root,
            either_versionfile_or_versionscript=True,
        )

        # In hatchling, versionfile-wheel setting doesn't get used.
        # Instead, the versionfile-sdist needs to be used to locate the build _version.py file.
        versionfile_sdist: Path | None = get_toml_value(
            pyproject_toml,
            ["tool", "version-pioneer", "versionfile-sdist"],
            return_path_object=True,
        )
        if versionfile_sdist is None:
            print("No versionfile-sdist specified in pyproject.toml")
            print("Skipping writing a constant version file")
            return
        else:
            self.temp_version_file = tempfile.NamedTemporaryFile(mode="w", delete=True)  # noqa: SIM115
            version_dict = exec_version_script(versionscript)
            self.temp_version_file.write(
                convert_version_dict(version_dict, output_format="python")
            )
            self.temp_version_file.flush()

            # make it executable
            versionfile_build_temp = Path(self.temp_version_file.name)
            versionfile_build_temp.chmod(
                versionfile_build_temp.stat().st_mode | stat.S_IEXEC
            )

            build_data["force_include"][self.temp_version_file.name] = versionfile_sdist

    def finalize(
        self,
        version: str,
        build_data: dict[str, Any],
        artifact_path: str,
    ) -> None:
        if self.temp_version_file is not None:
            # Delete the temporary version file
            self.temp_version_file.close()



import textwrap
from pathlib import Path

import pytest

from version_pioneer.template import NO_VENDOR_VERSIONSCRIPT
from version_pioneer.utils.build import build_project
from version_pioneer.utils.version_script import exec_version_script


def test_no_vendor_hatchling_parentdir_prefix_from_url(
    plugin_wheel: Path, tmp_path: Path, monkeypatch: pytest.MonkeyPatch
):
    """
    Using default NO_VENDOR_VERSIONSCRIPT template, version should be inferred from parent directory.
    """
    project_dir = tmp_path / "no_vendor_hatchling_project-0.7.0"
    project_dir.mkdir(parents=True, exist_ok=True)
    monkeypatch.chdir(project_dir)

    pyp = project_dir / "pyproject.toml"

    pyp.write_text(
        textwrap.dedent(f"""
            [build-system]
            requires = ["hatchling", "version-pioneer @ {plugin_wheel.as_uri()}"]
            build-backend = "hatchling.build"

            [tool.hatch.version]
            source = "version-pioneer"

            [tool.hatch.build.hooks.version-pioneer]

            [tool.version-pioneer]
            versionscript = "src/my_app/_version.py"
            versionfile-sdist = "src/my_app/_version.py"
            versionfile-wheel = "my_app/_version.py"

            [project]
            name = "kkkkkkk"  # <- NOT HERE
            dynamic = ["version"]
            requires-python = ">=3.8"

            [project.urls]
            source = "https://github.com/username/no_vendor_hatchling_project.git"  # <- HERE
        """)
    )

    (project_dir / "src" / "my_app").mkdir(parents=True, exist_ok=True)
    (project_dir / "src" / "my_app" / "_version.py").write_text(NO_VENDOR_VERSIONSCRIPT)

    version_dict = exec_version_script(project_dir / "src" / "my_app" / "_version.py")
    assert version_dict["version"] == "0.7.0"

    out, built_paths = build_project()
    assert len(built_paths) == 2
    assert "Successfully built dist/kkkkkkk-0.7.0-py3-none-any.whl" in out
    assert "Successfully built dist/kkkkkkk-0.7.0.tar.gz" in out


def test_no_vendor_hatchling_parentdir_prefix_from_project_name(
    plugin_wheel: Path, tmp_path: Path, monkeypatch: pytest.MonkeyPatch
):
    """
    Using default NO_VENDOR_VERSIONSCRIPT template, version should be inferred from parent directory.
    """
    project_dir = tmp_path / "infer-from-project-name-0.0.5"
    project_dir.mkdir(parents=True, exist_ok=True)
    monkeypatch.chdir(project_dir)

    pyp = project_dir / "pyproject.toml"

    pyp.write_text(
        textwrap.dedent(f"""
            [build-system]
            requires = ["hatchling", "version-pioneer @ {plugin_wheel.as_uri()}"]
            build-backend = "hatchling.build"

            [tool.hatch.version]
            source = "version-pioneer"

            [tool.hatch.build.hooks.version-pioneer]

            [tool.version-pioneer]
            versionscript = "src/my_app/_version.py"
            versionfile-sdist = "src/my_app/_version.py"
            versionfile-wheel = "my_app/_version.py"

            [project]
            name = "infer-from-project-name"  # <- HERE
            dynamic = ["version"]
            requires-python = ">=3.8"

            # [project.urls] does not exist
        """)
    )

    (project_dir / "src" / "my_app").mkdir(parents=True, exist_ok=True)
    (project_dir / "src" / "my_app" / "_version.py").write_text(NO_VENDOR_VERSIONSCRIPT)

    version_dict = exec_version_script(project_dir / "src" / "my_app" / "_version.py")
    assert version_dict["version"] == "0.0.5"

    out, built_paths = build_project()
    assert len(built_paths) == 2
    assert (
        "Successfully built dist/infer_from_project_name-0.0.5-py3-none-any.whl" in out
    )
    assert "Successfully built dist/infer_from_project_name-0.0.5.tar.gz" in out



"""
This module should be able to be imported from anywhere without any dependencies, because it's used in setup.
"""

import textwrap

EXEC_OUTPUT_PYTHON = textwrap.dedent(
    """
    #!/usr/bin/env python3
    # THIS "versionfile" IS GENERATED BY version-pioneer-{version_pioneer_version}
    # by evaluating the original versionscript and storing the computed versions as a constant.

    def get_version_dict():
        return {version_dict}


    if __name__ == "__main__":
        import json

        print(json.dumps(get_version_dict()))
    """
).strip()


SETUP_PY = textwrap.dedent("""
    from setuptools import setup
    from version_pioneer.build.setuptools import get_version, get_cmdclass

    setup(
        version=get_version(),
        cmdclass=get_cmdclass(),
    )
""").strip()


INIT_PY = textwrap.dedent("""
    from ._version import get_version_dict

    __version__ = get_version_dict()["version"]
""").strip()


NO_VENDOR_VERSIONSCRIPT = textwrap.dedent("""
    #!/usr/bin/env python3
    from pathlib import Path

    from version_pioneer.api import get_version_dict_wo_exec


    def get_version_dict():
        # NOTE: during installation, __file__ is not defined
        # When installed in editable mode, __file__ is defined
        # When installed in standard mode (when built), this file is replaced to a compiled versionfile.
        if "__file__" in globals():
            cwd = Path(__file__).parent
        else:
            cwd = Path.cwd()

        return get_version_dict_wo_exec(
            cwd=cwd,
            style="pep440",
            tag_prefix="v",
        )


    if __name__ == "__main__":
        import json

        print(json.dumps(get_version_dict()))
""").strip()



"""
Once the package is set up, run the following commands to build the package.

It should be compatible with multiple build backends.
"""

from __future__ import annotations

import logging
import os
import subprocess
from pathlib import Path
from shutil import rmtree

from tests.utils import (
    get_dynamic_version,
    verify_resolved_versionfile,
)
from version_pioneer.api import get_version_script_core_code
from version_pioneer.utils.build import build_project, unpack_wheel
from version_pioneer.utils.version_script import (
    exec_version_script_code,
)

logger = logging.getLogger(__name__)


def assert_build_and_version_persistence(project_dir: Path):
    """
    Build a fake project end-to-end and verify wheel contents.

    First, do it with tag v0.1.0, then with a commit, then with unstaged changes (dirty).
    """
    dynamic_version = get_dynamic_version(project_dir)

    build_project(cwd=project_dir)

    whl = project_dir / "dist" / "my_app-0.1.0-py3-none-any.whl"

    assert whl.exists(), f"Build did not produce a correctly named wheel. Found: {os.listdir(project_dir / 'dist')}"

    unpack_wheel(whl)

    resolved_versionfile = (
        project_dir / "my_app-0.1.0" / "my_app" / "_version.py"
    ).read_text()
    verify_resolved_versionfile(resolved_versionfile)

    # actually evaluate the version
    logger.info(f"Resolved _version.py code: {resolved_versionfile}")
    version_after_tag: str = exec_version_script_code(resolved_versionfile)["version"]
    logger.info(f"Version after tag: {version_after_tag}")

    assert version_after_tag == "0.1.0"
    assert version_after_tag == dynamic_version

    #############################################
    # the second build will have a different version.
    rmtree(project_dir / "dist")
    subprocess.run(["git", "add", "."], cwd=project_dir, check=True)
    subprocess.run(
        ["git", "commit", "--allow-empty", "-am", "Second commit"],
        cwd=project_dir,
        check=True,
    )

    ps = subprocess.run(
        ["git", "status"],
        cwd=project_dir,
        check=True,
        capture_output=True,
        text=True,
    )
    logger.info(ps.stdout)

    dynamic_version = get_dynamic_version(project_dir)
    logger.info(f"Version after one commit (dynamic): {dynamic_version}")

    assert dynamic_version != "0.1.0"
    assert dynamic_version.startswith("0.1.0+1.g")
    assert not dynamic_version.endswith(".dirty")

    build_project(cwd=project_dir)

    # whls = list((new_hatchling_project / "dist").glob("*.whl"))
    # logger.info(f"Found wheels: {whls}")
    whl = project_dir / "dist" / f"my_app-{dynamic_version}-py3-none-any.whl"
    # ps = subprocess.run(
    #     ["git", "status"],
    #     cwd=project_dir,
    #     check=True,
    #     capture_output=True,
    #     text=True,
    # )
    # logger.info(ps.stdout)
    assert whl.exists(), f"Build did not produce a correctly named wheel. Found: {os.listdir(project_dir / 'dist')}"

    unpack_wheel(whl)

    resolved_versionfile = (
        project_dir / f"my_app-{dynamic_version}" / "my_app" / "_version.py"
    ).read_text()
    verify_resolved_versionfile(resolved_versionfile)

    # actually evaluate the version
    version_after_commit_resolved = exec_version_script_code(resolved_versionfile)[
        "version"
    ]
    logger.info(f"Version after commit (resolved): {version_after_commit_resolved}")

    assert dynamic_version == version_after_commit_resolved

    #############################################
    # modify a file and see if .dirty is appended
    # only unstaged changes count, and not a new file. So we remove what we added earlier.
    rmtree(project_dir / "my_app-0.1.0")
    (project_dir / "aaa.txt").touch()
    subprocess.run(["git", "add", "."], cwd=project_dir, check=True)

    # ps = subprocess.run(
    #     ["git", "status"],
    #     cwd=project_dir,
    #     check=True,
    #     capture_output=True,
    #     text=True,
    # )
    # logger.info(ps.stdout)

    dynamic_version = get_dynamic_version(project_dir)
    logger.info(
        f"Version after one commit and unstaged changes (dynamic): {dynamic_version}"
    )

    assert dynamic_version != "0.1.0"
    assert dynamic_version.startswith("0.1.0+1.g")
    assert dynamic_version.endswith(".dirty")

    build_project(cwd=project_dir)

    # whls = list((new_hatchling_project / "dist").glob("*.whl"))
    # logger.info(f"Found wheels: {whls}")
    whl = project_dir / "dist" / f"my_app-{dynamic_version}-py3-none-any.whl"
    assert whl.exists(), f"Build did not produce a correctly named wheel. Found: {os.listdir(project_dir / 'dist')}"

    unpack_wheel(whl)

    resolved_versionfile = (
        project_dir / f"my_app-{dynamic_version}" / "my_app" / "_version.py"
    ).read_text()
    verify_resolved_versionfile(resolved_versionfile)

    # actually evaluate the version
    version_after_commit_resolved = exec_version_script_code(resolved_versionfile)[
        "version"
    ]
    logger.info(
        f"Version after commit and unstaged changes (resolved): {version_after_commit_resolved}"
    )

    assert dynamic_version == version_after_commit_resolved


def check_no_versionfile_output(*, cwd: Path, mode: str = "both", version="0.1.1"):
    """
    Check when versionfile-sdist or versionfile-wheel is not set.

    Note:
        Must be used with xfail(raise=VersionScriptResolutionError).

    Assume dist/ exists and contains the built files.
    """
    if mode not in ("sdist", "wheel", "both"):
        raise ValueError(f"Invalid mode: {mode}")
    if mode in ("sdist", "both"):
        sdist = cwd / "dist" / f"my_app-{version}.tar.gz"
        assert sdist.exists()
        subprocess.run(["tar", "xzf", sdist], cwd=cwd / "dist", check=True)
        unresolved_versionscript = (
            cwd / "dist" / f"my_app-{version}" / "src" / "my_app" / "_version.py"
        ).read_text()
        assert unresolved_versionscript == get_version_script_core_code()
        verify_resolved_versionfile(unresolved_versionscript)  # expected to fail
    if mode in ("wheel", "both"):
        # logger.info(list((cwd / "dist").glob("*")))
        whl = cwd / "dist" / f"my_app-{version}-py3-none-any.whl"

        assert whl.exists()

        unpack_wheel(whl, dest_dir=cwd / "dist")

        unresolved_versionscript = (
            cwd / "dist" / f"my_app-{version}" / "my_app" / "_version.py"
        ).read_text()
        assert unresolved_versionscript == get_version_script_core_code()
        verify_resolved_versionfile(unresolved_versionscript)  # expected to fail
    rmtree(cwd / "dist")



from __future__ import annotations

import inspect
import json
import logging
import os
import sys
from collections.abc import Sequence
from datetime import datetime, timezone
from functools import lru_cache
from importlib.metadata import Distribution, PackageNotFoundError
from os import PathLike
from pathlib import Path

from ._version import get_version_dict as _get_version_dict

__version__ = _get_version_dict()["version"]

logger = logging.getLogger(__name__)

APP_NAME = __name__
APP_NAME_UPPER = APP_NAME.upper()
PACKAGE_NAME = APP_NAME.replace("_", "-")


def setup_logging(
    console_level: int | str = logging.INFO,
    log_dir: str | PathLike | None = None,
    output_files: Sequence[str] = (
        "{date:%Y%m%d-%H%M%S}-{name}-{levelname}-{version}.log",
    ),
    file_levels: Sequence[int] = (logging.INFO,),
    *,
    log_init_messages: bool = True,
    console_formatter: logging.Formatter | None = None,
    file_formatter: logging.Formatter | None = None,
):
    r"""
    Setup logging with RichHandler and FileHandler.

    You should call this function at the beginning of your script.

    Args:
        console_level: Logging level for console. Defaults to INFO or env var {APP_NAME_UPPER}_LOG_LEVEL.
        log_dir: Directory to save log files. If None, only console logging is enabled. Usually set to LOG_DIR.
        output_files: List of output file paths, relative to log_dir. Only applies if log_dir is not None.
        file_levels: List of logging levels for each output file. Only applies if log_dir is not None.
        log_init_messages: Whether to log the initialisation messages.
    """
    try:
        from rich.console import Console
        from rich.logging import RichHandler
        from rich.theme import Theme
    except ModuleNotFoundError:
        print("⚠️ CLI dependencies are not installed.")  # noqa: T201
        print(  # noqa: T201
            "Please install Version-Pioneer with `pip install 'version-pioneer[cli]'`."
        )
        print("or even better, `uv tool install 'version-pioneer[cli]'`.")  # noqa: T201

        sys.exit(1)

    _console = Console(
        theme=Theme(
            {
                "logging.level.error": "bold red blink",
                "logging.level.critical": "red blink",
                "logging.level.warning": "yellow",
                "logging.level.success": "green",
            }
        )
    )

    @lru_cache
    def pkg_is_editable():
        try:
            direct_url = Distribution.from_name(PACKAGE_NAME).read_text(
                "direct_url.json"
            )
        except PackageNotFoundError:
            # Not installed?
            return False

        if direct_url is None:
            # package is not installed at all
            return False
        return json.loads(direct_url).get("dir_info", {}).get("editable", False)

    # NOTE: The value is None if you haven't installed with `pip install -e .` (development mode).
    # We make it None to discourage the use of this path. Only use for development.
    if pkg_is_editable():
        PROJECT_DIR = Path(__file__).parent.parent.parent  # noqa: N806
    else:
        PROJECT_DIR = None  # noqa: N806

    assert len(output_files) == len(
        file_levels
    ), "output_files and file_levels must have the same length"

    if log_dir is None:
        output_files = []
        file_levels = []
    else:
        log_dir = Path(log_dir)

    # NOTE: Initialise with NOTSET level and null device, and add stream handler separately.
    # This way, the root logging level is NOTSET (log all), and we can customise each handler's behaviour.
    # If we set the level during the initialisation, it will affect to ALL streams,
    # so the file stream cannot be more verbose (lower level) than the console stream.
    logging.basicConfig(
        format="",
        level=logging.NOTSET,
        stream=open(os.devnull, "w"),  # noqa: SIM115
    )

    # If you want to suppress logs from other modules, set their level to WARNING or higher
    # logging.getLogger('slowfast.utils.checkpoint').setLevel(logging.WARNING)

    console_handler = RichHandler(
        level=console_level,
        show_time=True,
        show_level=True,
        show_path=True,
        rich_tracebacks=True,
        tracebacks_show_locals=True,
        console=_console,
    )

    if console_formatter is None:
        console_format = logging.Formatter(
            fmt="%(name)s - %(message)s",
            datefmt="%H:%M:%S",
        )
    else:
        console_format = console_formatter
    console_handler.setFormatter(console_format)

    if file_formatter is None:
        f_format = logging.Formatter(
            fmt="%(asctime)s - %(name)s: %(lineno)4d - %(levelname)s - %(message)s",
            datefmt="%y/%m/%d %H:%M:%S",
        )
    else:
        f_format = file_formatter

    function_caller_module = inspect.getmodule(inspect.stack()[1][0])
    if function_caller_module is None:
        name_or_path = "unknown"
    elif function_caller_module.__name__ == "__main__":
        if function_caller_module.__file__ is None:
            name_or_path = function_caller_module.__name__
        elif PROJECT_DIR is not None:
            # Called from files in the project directory.
            # Instead of using the __name__ == "__main__", infer the module name from the file path.
            name_or_path = function_caller_module.__file__.replace(
                str(PROJECT_DIR) + "/", ""
            ).replace("/", ".")
            # Remove .py extension
            name_or_path = Path(name_or_path).with_suffix("")
        else:
            # Called from somewhere outside the project directory.
            # Use the script name, like "script.py"
            name_or_path = Path(function_caller_module.__file__).name
    else:
        name_or_path = function_caller_module.__name__

    log_path_map = {
        "name": name_or_path,
        "version": __version__,
        "date": datetime.now(timezone.utc),
    }

    root_logger = logging.getLogger()
    root_logger.addHandler(console_handler)

    if log_init_messages:
        logger.info(f"Running with {PACKAGE_NAME} {__version__}")

    if log_dir is not None:
        log_paths: list[Path] = []
        for output_file, file_level in zip(output_files, file_levels):
            log_path_map["levelname"] = logging._levelToName[file_level]
            log_path = log_dir / output_file.format_map(log_path_map)
            log_path.parent.mkdir(parents=True, exist_ok=True)
            log_paths.append(log_path)

            f_handler = logging.FileHandler(log_path)
            f_handler.setLevel(file_level)
            f_handler.setFormatter(f_format)

            # Add handlers to the logger
            root_logger.addHandler(f_handler)

        if log_init_messages:
            for log_path in log_paths:
                logger.info(f"Logging to {log_path}")



import logging
import subprocess
import textwrap
from pathlib import Path

import pytest

from version_pioneer.api import build_consistency_test
from version_pioneer.utils.build import build_project

from .build_pipelines import (
    assert_build_and_version_persistence,
    check_no_versionfile_output,
)
from .utils import (
    VersionScriptResolutionError,
)

logger = logging.getLogger(__name__)


def test_build_consistency(new_pdm_project: Path):
    # Reset the project to a known state.
    subprocess.run(["git", "stash", "--all"], cwd=new_pdm_project, check=True)
    subprocess.run(["git", "checkout", "v0.1.0"], cwd=new_pdm_project, check=True)
    build_consistency_test(project_dir=new_pdm_project)


def test_build_version(new_pdm_project: Path):
    # Reset the project to a known state.
    subprocess.run(["git", "stash", "--all"], cwd=new_pdm_project, check=True)
    subprocess.run(["git", "checkout", "v0.1.0"], cwd=new_pdm_project, check=True)
    assert_build_and_version_persistence(new_pdm_project)


def test_different_versionfile(new_pdm_project: Path, plugin_wheel: Path):
    # Reset the project to a known state.
    subprocess.run(["git", "stash", "--all"], cwd=new_pdm_project, check=True)
    subprocess.run(["git", "checkout", "v0.1.0"], cwd=new_pdm_project, check=True)

    pyp = new_pdm_project / "pyproject.toml"

    pyp.write_text(
        textwrap.dedent(f"""
            [build-system]
            requires = ["pdm-backend", "version-pioneer @ {plugin_wheel.as_uri()}"]
            build-backend = "pdm.backend"

            [tool.version-pioneer]
            versionscript = "src/my_app/_version.py"
            versionfile-sdist = "src/my_app/_versionfile.py"
            versionfile-wheel = "my_app/_versionfile.py"

            [project]
            name = "my-app"
            dynamic = ["version"]
            requires-python = ">=3.8"
        """),
    )

    subprocess.run(["git", "add", "."], check=True)
    subprocess.run(["git", "commit", "-m", "Second commit"], check=True)
    subprocess.run(["git", "tag", "v0.1.1"], check=True)

    build_consistency_test(project_dir=new_pdm_project, expected_version="0.1.1")


def test_invalid_config(new_pdm_project: Path, plugin_wheel: Path):
    """
    Missing config makes the build fail with a meaningful error message.
    """
    # Reset the project to a known state.
    subprocess.run(["git", "stash", "--all"], cwd=new_pdm_project, check=True)
    subprocess.run(["git", "checkout", "v0.1.0"], cwd=new_pdm_project, check=True)

    pyp = new_pdm_project / "pyproject.toml"

    # If we leave out the config for good, the plugin doesn't get activated.
    pyp.write_text(
        textwrap.dedent(f"""
            [build-system]
            requires = ["pdm-backend", "version-pioneer @ {plugin_wheel.as_uri()}"]
            build-backend = "pdm.backend"

            [tool.version-pioneer]
            # MISSING CONFIGURATION

            [project]
            name = "my-app"
            dynamic = ["version"]
            requires-python = ">=3.8"
        """),
    )

    err, _ = build_project(check=False)

    assert (
        "KeyError: 'Missing key tool.version-pioneer.versionscript in pyproject.toml'"
        in err
    ), err

    pyp.write_text(
        textwrap.dedent(f"""
            [build-system]
            requires = ["pdm-backend", "version-pioneer @ {plugin_wheel.as_uri()}"]
            build-backend = "pdm.backend"

            [tool.version-pioneer]
            # versionscript = "src/my_app/_version.py"
            versionfile-sdist = "src/my_app/_version.py"
            versionfile-wheel = "my_app/_version.py"

            [project]
            name = "my-app"
            dynamic = ["version"]
            requires-python = ">=3.8"
        """),
    )

    err, _ = build_project(check=False)

    assert (
        "KeyError: 'Missing key tool.version-pioneer.versionscript in pyproject.toml'"
        in err
    ), err


@pytest.mark.xfail(raises=VersionScriptResolutionError)
def test_no_versionfile_sdist(new_pdm_project: Path, plugin_wheel: Path):
    """
    If versionfile-sdist is not configured, the build does NOT FAIL but the _version.py file is not updated.
    """
    # Reset the project to a known state.
    subprocess.run(["git", "stash", "--all"], cwd=new_pdm_project, check=True)
    subprocess.run(["git", "checkout", "v0.1.0"], cwd=new_pdm_project, check=True)

    pyp = new_pdm_project / "pyproject.toml"

    pyp.write_text(
        textwrap.dedent(f"""
            [build-system]
            requires = ["pdm-backend", "version-pioneer @ {plugin_wheel.as_uri()}"]
            build-backend = "pdm.backend"

            [tool.version-pioneer]
            versionscript = "src/my_app/_version.py"
            # versionfile-sdist = "src/my_app/_version.py"
            # versionfile-wheel = "my_app/_version.py"

            [project]
            name = "my-app"
            dynamic = ["version"]
            requires-python = ">=3.8"
        """),
    )

    subprocess.run(["git", "add", "."], check=True)
    subprocess.run(["git", "commit", "-m", "Second commit"], check=True)
    subprocess.run(["git", "tag", "v0.1.1"], check=True)

    # NOTE: consistency check fails on PDM because when building both sdist and wheel at the same time,
    # the wheel is built from the sdist, and the sdist doesn't have git information because we do not change
    # the versionfile.
    # This is rather expected, because the user specifically asked to not update the versionfile.
    # This works in hatchling because they provide a way to execute the version script directly.
    # It's just a minor difference in how they work, which should be noted, but it's not a bug.
    # You just made a wrong sdist, because sdist has to be complete and self-contained.
    # assert_build_consistency(version="0.1.1", cwd=new_pdm_project)  # Fails here.

    # Instead, we build the sdist and wheel separately.
    build_project("--sdist", cwd=new_pdm_project)
    build_project("--wheel", cwd=new_pdm_project)

    check_no_versionfile_output(cwd=new_pdm_project)



# Allow print
# Allow many arguments
# Allow relative import from parent
# Allow using Optional
# ruff: noqa: T201 PLR0913 TID252 FA100

# NOTE: type | None only works in Python 3.10+ with typer, so we use Optional instead.

import sys

try:
    import rich
    import typer
    from rich.prompt import Confirm
    from rich.syntax import Syntax

    from .docstring import from_docstring
except ModuleNotFoundError:
    print("⚠️ CLI dependencies are not installed.")
    print("Please install Version-Pioneer with `pip install 'version-pioneer[cli]'`.")
    print("or even better, `uv tool install 'version-pioneer[cli]'`.")

    sys.exit(1)


from pathlib import Path
from typing import List, Optional

from version_pioneer.template import INIT_PY, NO_VENDOR_VERSIONSCRIPT, SETUP_PY
from version_pioneer.utils.diff import unidiff_output
from version_pioneer.utils.version_script import ResolutionFormat
from version_pioneer.versionscript import VersionStyle

if sys.version_info < (3, 9):
    from typing_extensions import Annotated
else:
    from typing import Annotated

app = typer.Typer(
    no_args_is_help=True,
    context_settings={"help_option_names": ["-h", "--help"]},
    help="🧗 Version-Pioneer: Dynamically manage project version with hatchling and pdm support.",
)


def version_callback(*, value: bool):
    if value:
        from .. import __version__

        print(__version__)
        raise typer.Exit


@app.callback()
def common(
    ctx: typer.Context,
    *,
    version: bool = typer.Option(
        None, "-v", "--version", callback=version_callback, help="Show version"
    ),
):
    pass


@app.command()
@from_docstring
def install(
    project_dir: Annotated[Optional[Path], typer.Argument()] = None,
    *,
    vendor: bool = True,
):
    """
    Add _version.py, modify __init__.py and maybe setup.py.

    Args:
        project_dir: The root or child directory of the project. Default is cwd.
        vendor: Install the full versionscript. --no-vendor to import from version_pioneer.
    """
    from version_pioneer.api import get_version_script_core_code
    from version_pioneer.utils.toml import (
        find_pyproject_toml,
        get_toml_value,
        load_toml,
    )

    def _write_file_with_diff_confirm(file: Path, content: str):
        if file.exists():
            existing_content = file.read_text()
            if existing_content.strip() == content.strip():
                rich.print(f"[green]File already exists:[/green] {file} (no changes)")
                sys.exit(2)

            unified_diff = unidiff_output(existing_content, content)
            rich.print(
                Syntax(unified_diff, "diff", line_numbers=True, theme="lightbulb")
            )
            print()

            confirm = Confirm.ask(
                f"File [green]{file}[/green] already exists. [red]Overwrite?[/red]",
                default=False,
            )
            if not confirm:
                rich.print("[red]Aborted.[/red]")
                sys.exit(1)

        file.write_text(content)
        rich.print(f"[green]File written:[/green] {file}")

    pyproject_toml_file = find_pyproject_toml(project_dir)
    pyproject_toml = load_toml(pyproject_toml_file)

    project_dir = pyproject_toml_file.parent
    version_script_file = project_dir / Path(
        get_toml_value(
            pyproject_toml,
            ["tool", "version-pioneer", "versionscript"],
            raise_error=True,
        )
    )

    if vendor:
        _write_file_with_diff_confirm(
            version_script_file, get_version_script_core_code()
        )
    else:
        _write_file_with_diff_confirm(version_script_file, NO_VENDOR_VERSIONSCRIPT)

    # Modify __init__.py
    init_py_file = version_script_file.parent / "__init__.py"
    if not init_py_file.exists():
        init_py_file.write_text(INIT_PY)
        rich.print(f"[green]{init_py_file} added with content:[/green]")
        print(INIT_PY)
    else:
        init_py_content = init_py_file.read_text()
        init_py_template_lines = [line for line in INIT_PY.splitlines() if line.strip()]
        # if all lines exists in the init_py_content
        if all(line in init_py_content for line in init_py_template_lines):
            print("__init__.py already configured. Not modifying.")
        else:
            init_py_file.write_text(INIT_PY + "\n\n" + init_py_content)
            rich.print(f"[green]{init_py_file} modified with[/green]")
            print(INIT_PY)
            rich.print("[green]at the top![/green]")

    # Using setuptools.build_meta backend?
    try:
        build_backend = get_toml_value(
            pyproject_toml, ["build-system", "build-backend"], raise_error=True
        )
    except KeyError:
        confirm = Confirm.ask(
            "Are you using setuptools.build_meta backend? Install setup.py?",
            default=False,
        )

        if confirm:
            build_backend = "setuptools.build_meta"
        else:
            build_backend = None

    if build_backend is not None and build_backend == "setuptools.build_meta":
        # install setup.py
        setup_py_file = project_dir / "setup.py"
        _write_file_with_diff_confirm(setup_py_file, SETUP_PY)

    rich.print("[green]Installation completed![/green]")


@app.command()
def print_versionscript_code():
    """Print the content of versionscript.py file (for manual installation)."""
    from version_pioneer.api import get_version_script_core_code

    print(get_version_script_core_code())


@app.command()
def exec_version_script(
    project_dir_or_version_script_file: Annotated[
        Optional[Path], typer.Argument()
    ] = None,
    output_format: ResolutionFormat = ResolutionFormat.version_string,
):
    """Resolve the _version.py file for build, and print the content."""
    from version_pioneer.api import exec_version_script_and_convert

    print(
        exec_version_script_and_convert(
            project_dir_or_version_script_file, output_format=output_format
        )
    )


@app.command()
@from_docstring
def get_version_wo_exec(
    project_dir: Annotated[Optional[Path], typer.Argument()] = None,
    *,
    style: VersionStyle = VersionStyle.pep440,
    tag_prefix: str = "v",
    parentdir_prefix: Optional[str] = None,
    output_format: ResolutionFormat = ResolutionFormat.version_string,
    verbose: bool = False,
):
    """
    WITHOUT evaluating the _version.py file, get version from VCS with built-in Version-Pioneer logic.

    Useful when you don't need to customise the _version.py file, and you work in non-Python projects
    so you don't care about re-evaluating the version file.

    Args:
        project_dir: The root or child directory of the project. Default is cwd.
        parentdir_prefix: The prefix of the parent directory. (e.g. {github_repo_name}-)
    """
    from version_pioneer.api import get_version_wo_exec_and_convert

    print(
        get_version_wo_exec_and_convert(
            project_dir,
            style=style,
            tag_prefix=tag_prefix,
            parentdir_prefix=parentdir_prefix,
            output_format=output_format,
            verbose=verbose,
        )
    )


@app.command()
@from_docstring
def build_consistency_test(
    project_dir: Annotated[Optional[Path], typer.Argument()] = None,
    *,
    delete_temp_dir: bool = True,
    expected_version: Optional[str] = None,
    test_chaining: bool = True,
    ignore_patterns: Annotated[List[str], typer.Option("--ignore-pattern", "-i")] = [  # noqa: B006
        "*.egg-info/SOURCES.txt"
    ],
):
    """
    Check if builds are consistent with sdist, wheel, both, sdist -> sdist.

    Args:
        project_dir: The root or child directory of the project. Default is cwd.
        expected_version: Check if it builds to the expected version (without tag prefix).
        ignore_patterns: List of patterns to ignore when seeing diff of directory.
        test_chaining: Test sdist -> sdist chaining.
            Note that some build backends may produce different results.
            For example, setuptools produces setup.cfg in the first build,
            so the second result will have one more file in the SOURCES.txt list.
    """
    from version_pioneer import setup_logging
    from version_pioneer.api import build_consistency_test

    setup_logging()
    build_consistency_test(
        project_dir,
        delete_temp_dir=delete_temp_dir,
        test_chaining=test_chaining,
        expected_version=expected_version,
        ignore_patterns=ignore_patterns,
    )


def main():
    app()


if __name__ == "__main__":
    main()
